"""
transform: functions to convert data between composite structures
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2021, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    
          
To Do:
    Complete functions that are presently not implemented.
    Add in 'dispatcher' decorators to the converter functions.
    Fix 'adjacency_to_matrix' - it currently returns all zeros in the matrix.
    
"""
from __future__ import annotations
import abc
import collections
from collections.abc import Collection
import dataclasses
from typing import Any, TYPE_CHECKING

import more_itertools

from . import check
   
if TYPE_CHECKING:
    from . import composites
    from . import graph
    from . import hybrid
    from . import tree
    
""" Converters """

# @amos.dispatcher 
def to_adjacency(item: Any) -> graph.Adjacency:
    """Converts 'item' to an Adjacency.
    
    Args:
        item (Any): item to convert to an Adjacency.

    Raises:
        TypeError: if 'item' is a type that is not registered with the 
        dispatcher.

    Returns:
        graph.Adjacency: derived from 'item'.

    """
    if check.is_adjacency(item = item):
        return item
    else:
        raise TypeError(
            f'item cannot be converted because it is an unsupported type: '
            f'{type(item).__name__}')

# @to_adjacency.register # type: ignore
def edges_to_adjacency(item: graph.Edges) -> graph.Adjacency:
    """Converts 'item' to an Adjacency.

    Args:
        item (graph.Edges): item to convert to an Adjacency.

    Returns:
        graph.Adjacency: derived from 'item'.

    """
    adjacency = collections.defaultdict(set)
    for edge_pair in item:
        if edge_pair[0] not in adjacency:
            adjacency[edge_pair[0]] = {edge_pair[1]}
        else:
            adjacency[edge_pair[0]].add(edge_pair[1])
        if edge_pair[1] not in adjacency:
            adjacency[edge_pair[1]] = set()
    return adjacency

# @to_adjacency.register # type: ignore 
def matrix_to_adjacency(item: graph.Matrix) -> graph.Adjacency:
    """Converts 'item' to an Adjacency.

    Args:
        item (graph.Matrix): item to convert to an Adjacency.

    Returns:
        graph.Adjacency: derived from 'item'.

    """  
    matrix = item[0]
    names = item[1]
    name_mapping = dict(zip(range(len(matrix)), names))
    raw_adjacency = {
        i: [j for j, adjacent in enumerate(row) if adjacent] 
        for i, row in enumerate(matrix)}
    adjacency = collections.defaultdict(set)
    for key, value in raw_adjacency.items():
        new_key = name_mapping[key]
        new_values = set()
        for edge in value:
            new_values.add(name_mapping[edge])
        adjacency[new_key] = new_values
    return adjacency

# @to_adjacency.register # type: ignore 
def pipeline_to_adjacency(item: hybrid.Pipeline) -> graph.Adjacency:
    """Converts 'item' to an Adjacency.

    Args:
        item (hybrid.Pipeline): item to convert to an Adjacency.

    Returns:
        graph.Adjacency: derived from 'item'.

    """
    if not isinstance(item, (Collection)) or isinstance(item, str):
        item = [item]
    adjacency = collections.defaultdict(set)
    if len(item) == 1:
        adjacency.update({item: set()})
    else:
        edges = more_itertools.windowed(item, 2)
        for edge_pair in edges:
            if edge_pair[0] in adjacency:
                adjacency[edge_pair[0]].add(edge_pair[1])
            else:
                adjacency[edge_pair[0]] = {edge_pair[1]}
    return adjacency

# @to_adjacency.register # type: ignore 
def pipelines_to_adjacency(item: hybrid.Pipelines) -> graph.Adjacency:
    """Converts 'item' to an Adjacency.

    Args:
        item (hybrid.Pipelines): item to convert to an Adjacency.

    Returns:
        graph.Adjacency: derived from 'item'.

    """
    adjacency = collections.defaultdict(set)
    for _, pipeline in item.items():
        pipe_adjacency = pipeline_to_adjacency(item = pipeline)
        for key, value in pipe_adjacency.items():
            if key in adjacency:
                for inner_value in value:
                    if inner_value not in adjacency:
                        adjacency[key].add(inner_value)
            else:
                adjacency[key] = value
    return adjacency

# @to_adjacency.register # type: ignore 
def tree_to_adjacency(item: tree.Tree) -> graph.Adjacency:
    """Converts 'item' to an Adjacency.

    Args:
        item (tree.Tree): item to convert to an Adjacency.

    Returns:
        graph.Adjacency: derived from 'item'.

    """
    raise NotImplementedError
             
# @to_adjacency.register # type: ignore 
def nodes_to_adjacency(item: composites.Nodes) -> graph.Adjacency:
    """Converts 'item' to an Adjacency.

    Args:
        item (composites.Nodes): item to convert to an Adjacency.

    Returns:
        graph.Adjacency: derived from 'item'.

    """
    adjacency = collections.defaultdict(set)
    return adjacency.update((k, set()) for k in item)

# @amos.dispatcher   
def to_edges(item: Any) -> graph.Edges:
    """Converts 'item' to an Edges.
    
    Args:
        item (Any): item to convert to an Edges.

    Raises:
        TypeError: if 'item' is a type that is not registered.

    Returns:
        graph.Edges: derived from 'item'.

    """
    if check.is_edges(item = item):
        return item
    else:
        raise TypeError(
            f'item cannot be converted because it is an unsupported type: '
            f'{type(item).__name__}')
    
# @to_edges.register # type: ignore
def adjacency_to_edges(item: graph.Adjacency) -> graph.Edges:
    """Converts 'item' to an Edges.
    
    Args:
        item (graph.Adjacency): item to convert to an Edges.

    Returns:
        graph.Edges: derived from 'item'.

    """ 
    edges = []
    for node, connections in item.items():
        for connection in connections:
            edges.append(tuple([node, connection]))
    return tuple(edges)

# @amos.dispatcher   
def to_matrix(item: Any) -> graph.Matrix:
    """Converts 'item' to a Matrix.
    
    Args:
        item (Any): item to convert to a Matrix.

    Raises:
        TypeError: if 'item' is a type that is not registered.

    Returns:
        graph.Matrix: derived from 'item'.

    """
    if check.is_matrix(item = item):
        return item
    else:
        raise TypeError(
            f'item cannot be converted because it is an unsupported type: '
            f'{type(item).__name__}')

# @to_matrix.register # type: ignore 
def adjacency_to_matrix(item: graph.Adjacency) -> graph.Matrix:
    """Converts 'item' to a Matrix.
    
    Args:
        item (graph.Adjacency): item to convert to a Matrix.

    Returns:
        graph.Matrix: derived from 'item'.

    """ 
    names = list(item.keys())
    matrix = []
    for i in range(len(item)): 
        matrix.append([0] * len(item))
        for j in item[i]:
            matrix[i][j] = 1
    return tuple([matrix, names])    

# @amos.dispatcher   
def to_tree(item: Any) -> tree.Tree:
    """Converts 'item' to a Tree.
    
    Args:
        item (Any): item to convert to a Tree.

    Raises:
        TypeError: if 'item' is a type that is not registered.

    Returns:
        tree.Tree: derived from 'item'.

    """
    if check.is_tree(item = item):
        return item
    else:
        raise TypeError(
            f'item cannot be converted because it is an unsupported type: '
            f'{type(item).__name__}')

# @to_tree.register # type: ignore 
def matrix_to_tree(item: graph.Matrix) -> tree.Tree:
    """Converts 'item' to a Tree.
    
    Args:
        item (graph.Matrixy): item to convert to a Tree.

    Raises:
        TypeError: if 'item' is a type that is not registered.

    Returns:
        tree.Tree: derived from 'item'.

    """
    tree = {}
    for node in item:
        children = item[:]
        children.remove(node)
        tree[node] = matrix_to_tree(children)
    return tree

 


"""
trees: tree data structures 
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2021, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:

        
To Do:
    Fix 'breadth_first_search' algorithm for speed.
   
"""
from __future__ import annotations
import abc
from collections.abc import MutableSequence, Sequence
import dataclasses
from typing import Any, Callable, ClassVar, Optional, Type, TypeVar, Union

from ..base import sequences
from . import composites


""" Type Aliases """

Changer: Type[Any] = Callable[[composites.Node], None]
Finder: Type[Any] = Callable[[composites.Node], Optional[composites.Node]]


@dataclasses.dataclass # type: ignore
class Tree(sequences.Hybrid, composites.Composite, abc.ABC):
    """composites class for an tree data structures.
    
    The Tree class uses a Hybrid instead of a linked list for storing children
    nodes to allow easier access of nodes further away from the root. For
    example, a user might use 'a_tree["big_branch"]["small_branch"]["a_leaf"]' 
    to access a desired node instead of 'a_tree[2][0][3]' (although the latter
    access technique is also supported).
    
    There are several differences between a Tree and a Graph in piles:
        1) Graphs are more flexible. Trees must have 1 root, are directed, and
            each node can have only 1 parent node.
        2) Edges are only implicit in a Tree whereas they are explicit in a 
            Graph. This allows for certain methods and functions surrounding
            iteration and traversal to be faster.
        3) As the size of the data structure increases, a Tree should use less
            memory because the data about relationships between nodes is not
            centrally maintained (as with an adjacency matrix). This decreases
            access time to non-consecutive nodes, but is more efficient for 
            larger data structures.
        
    Args:
        contents (MutableSequence[Node]): list of stored Tree or other 
            Node instances. Defaults to an empty list.
        name (Optional[str]): name of Tree node which should match a parent 
            tree's key name corresponding to this Tree node. All nodes in a Tree
            must have unique names. The name is used to make all Tree nodes 
            hashable and capable of quick comparison. Defaults to None, but it
            should not be left as None when added to a Tree.
        parent (Optional[Tree]): parent Tree, if any. Defaults to None.
  
    """
    contents: MutableSequence[composites.Node] = dataclasses.field(
        default_factory = list)
    name: Optional[str] = None
    parent: Optional[Tree] = None 
    
    """ Properties """
    
    @property
    def children(self) -> MutableSequence[composites.Node]:
        return self.contents
    
    @children.setter
    def children(self, value: MutableSequence[composites.Node]) -> None:
        self.contents = value
        return
    
    @children.deleter
    def children(self, key: composites.Node) -> None:
        self.delete(key)
        return
           
    """ Dunder Methods """

    def __add__(self, other: composites.Composite) -> None:
        """Adds 'other' to the stored tree using the 'append' method.

        Args:
            other (composites.Composite): another Tree, an adjacency list, an 
                edge list, an adjacency matrix, or one or more nodes.
            
        """
        self.append(item = other)     
        return 

    def __radd__(self, other: composites.Composite) -> None:
        """Adds 'other' to the stored tree using the 'prepend' method.

        Args:
            other (composites.Composite): another Tree, an adjacency list, an 
                edge list, an adjacency matrix, or one or more nodes.
            
        """
        self.prepend(item = other)     
        return 

    def __missing__(self) -> dict[str, Tree]:
        """[summary]

        Returns:
            dict[str, Tree]: [description]
            
        """
        return self.__class__()
    
    def __hash__(self) -> int:
        """[summary]

        Returns:
            int: [description]
            
        """
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        """[summary]

        Args:
            other (Any): [description]

        Returns:
            bool: [description]
            
        """
        if hasattr(other, 'name'):
            return other.name == self.name
        else:
            return False
        
    def __ne__(self, other: Any) -> bool:
        """[summary]

        Args:
            other (Any): [description]

        Returns:
            bool: [description]
            
        """
        return not self.__eq__(other = other)


@dataclasses.dataclass # type: ignore
class Categorizer(Tree):
    """composites class for an tree data structures.
        
    Args:
        contents (MutableSequence[composites.Node]): list of stored Node 
            instances (including other Trees). Defaults to an empty list.
        name (Optional[str]): name of Tree node which should match a parent 
            tree's key name corresponding to this Tree node. All nodes in a Tree
            must have unique names. The name is used to make all Tree nodes 
            hashable and capable of quick comparison. Defaults to None, but it
            should not be left as None when added to a Tree.
        parent (Optional[Tree]): parent Tree, if any. Defaults to None.
        
    """
    contents: MutableSequence[composites.Node] = dataclasses.field(
        default_factory = list)
    name: Optional[str] = None
    parent: Optional[Tree] = None 
    
    """ Properties """
        
    @property
    def branches(self) -> list[Tree]:
        """Returns all stored Tree nodes in a list."""
        return self.nodes - self.leaves
    
    @property
    def children(self) -> dict[str, composites.Node]:
        """[summary]

        Returns:
            dict[str, composites.Node]: [description]
        """
        return self.contents
    
    @property
    def is_leaf(self) -> bool:
        """[summary]

        Returns:
            bool: [description]
        """
        return not self.children
    
    @property
    def is_root(self) -> bool:
        """[summary]

        Returns:
            bool: [description]
        """
        return self.parent is None
    
    @property
    def leaves(self) -> list[composites.Node]:
        """Returns all stored leaf nodes in a list."""
        matches = []
        for node in self.nodes:
            if not hasattr(node, 'is_leaf') or node.is_leaf:
                matches.append(node)
        return matches
     
    @property
    def nodes(self) -> list[composites.Node]:
        """Returns all stored nodes in a list."""
        return depth_first_search(tree = self.contents)

    @property
    def root(self) -> Tree:
        """
        """
        composites = [n.is_root for n in self.nodes]
        if len(composites) > 1:
            raise ValueError('The tree is broken - it has more than 1 root')
        elif len(composites) == 0:
            raise ValueError('The tree is broken - it has no root')
        else:
            return composites[0]
    
    """ Public Methods """
    
    def add(
        self, 
        item: Union[composites.Node, Sequence[composites.Node]],
        parent: Optional[str] = None) -> None:
        """Adds node(s) in item to 'contents'.
        
        In adding the node(s) to the stored tree, the 'parent' attribute for the
        node(s) is set to this Tree instance.

        Args:
            item (Union[composites.Node, Sequence[composites.Node]]): node(s) to 
                add to the 'contents' attribute.

        Raises:
            ValueError: if 'item' already is in the stored tree or if 'parent'
                is not in the tree.
                            
        """
        if parent is None:
            parent_node = self
        else:
            parent_node = self.get(item = parent)
        if parent_node is None:
            raise ValueError(
                f'Cannot add {item.name} because parent node {parent} is not '
                f'in the tree')
        if isinstance(item, Sequence) and not isinstance(item, str):
            for node in item:
                self.add(item = node)
        elif item in self.nodes:
            raise ValueError(
                f'Cannot add {item.name} because it is already in the tree')
        else:
            item.parent = parent_node
            parent_node.contents.append(item)
        return
    
    def find(self, finder: Finder, **kwargs: Any) -> Optional[composites.Node]:
        """Finds first matching node in Tree using 'finder'.

        Args:
            finder (Callable[[composites.Node], Optional[composites.Node]]): 
                function or other callable that returns a node if it meets 
                certain criteria or otherwise returns None.
            kwargs: keyword arguments to pass to 'finder' when examing each
                node.

        Returns:
            Optional[composites.Node]: matching Node or None if no matching node 
                is found.
            
        """                  
        for node in self.nodes:
            comparison = finder(self, **kwargs)
            if comparison:
                return node
        return None
            
    def find_add(
        self, 
        finder: Finder, 
        item: composites.Node, 
        **kwargs: Any) -> None:
        """Finds first matching node in Tree using 'finder'.

        Args:
            finder (Callable[[composites.Node], Optional[composites.Node]]): 
                function or other callable that returns a node if it meets 
                certain criteria or otherwise returns None.
            item (composites.Node): node to add to the 'contents' attribute of 
                the first node that meets criteria in 'finder'.
            kwargs: keyword arguments to pass to 'finder' when examing each
                node.

        Raises:
            ValueError: if no matching node is found by 'finder'.

        Returns:
            Optional[composites.Node]: matching Node or None if no matching node 
                is found.
            
        """  
        node = self.find(finder = finder, **kwargs)
        if node:
            node.add(item = item)
        else:
            raise ValueError(
                'item could not be added because no matching node was found by '
                'finder')
        return
    
    def find_all(self, finder: Finder, **kwargs: Any) -> list[composites.Node]:
        """Finds all matching nodes in Tree using 'finder'.

        Args:
            finder (Callable[[composites.Node], Optional[composites.Node]]): 
                function or other callable that returns a node if it meets 
                certain criteria or otherwise returns None.
            kwargs: keyword arguments to pass to 'finder' when examing each
                node.

        Returns:
            list[composites.Node]: matching nodes or an empty list if no 
                matching node is found.
            
        """              
        found = []     
        for node in self.nodes:
            comparison = finder(self, **kwargs)
            if comparison:
                found.append(node)
        return found
            
    def find_change(
        self, 
        finder: Finder, 
        changer: Changer, 
        **kwargs: Any) -> None:
        """Finds matching nodes in Tree using 'finder' and applies 'changer'.

        Args:
            finder (Callable[[composites.Node], Optional[composites.Node]]): 
                function or other callable that returns a node if it meets 
                certain criteria or otherwise returns None.
            changer (Callable[[composites.Node], None]): function or other 
                callable that modifies the found node.
            kwargs: keyword arguments to pass to 'finder' when examing each
                node.

        Raises:
            ValueError: if no matching node is found by 'finder'.
            
        """  
        nodes = self.find_all(finder = finder, **kwargs)
        if nodes:
            for node in nodes:
                changer(node)
        else:
            raise ValueError(
                'changer could not be applied because no matching node was '
                'found by finder')
        return
    
    def get(self, item: str) -> Optional[composites.Node]:
        """Finds first matching node in Tree match 'item'.

        Args:
            item (str): 

        Returns:
            Optional[composites.Node]: matching Node or None if no matching node 
                is found.
            
        """                  
        for node in self.nodes:
            if node.name == item:
                return node
        return self.__missing__()
                                    
    def walk(self, depth_first: bool = True) -> composites.Pipeline:
        """Returns all paths in tree from 'start' to 'stop'.
        
        Args:
            depth_first (bool): whether to search through the stored tree depth-
                first (True) or breadth_first (False). Defaults to True.
                
        """
        if depth_first:
            return depth_first_search(tree = self.contents)
        else:
            raise NotImplementedError(
                'breadth first search is not yet implemented')
            # return breadth_first_search(tree = self.contents)

    """ Dunder Methods """

    def __add__(self, other: composites.Composite) -> None:
        """Adds 'other' to the stored tree using the 'append' method.

        Args:
            other (composites.Composite): another Composite or supported
                raw data structure.
            
        """
        self.append(item = other)     
        return 

    def __radd__(self, other: composites.Composite) -> None:
        """Adds 'other' to the stored tree using the 'prepend' method.

        Args:
            other (composites.Composite): another Composite or supported
                raw data structure.
            
        """
        self.prepend(item = other)     
        return 

    def __missing__(self) -> dict[str, Tree]:
        """[summary]

        Returns:
            dict[str, Tree]: [description]
            
        """
        return {}
    
    def __hash__(self) -> int:
        """[summary]

        Returns:
            int: [description]
            
        """
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        """[summary]

        Args:
            other (Any): [description]

        Returns:
            bool: [description]
            
        """
        if hasattr(other, 'name'):
            return other.name == self.name
        else:
            return False
        
    def __ne__(self, other: Any) -> bool:
        """[summary]

        Args:
            other (Any): [description]

        Returns:
            bool: [description]
            
        """
        return not self.__eq__(other = other)


# def breadth_first_search(
#     tree: Tree, 
#     visited: Optional[list[Tree]] = None) -> composites.Pipeline:
#     """Returns a breadth first search path through 'tree'.

#     Args:
#         tree (Tree): tree to search.
#         visited (Optional[list[Tree]]): list of visited nodes. Defaults to None.

#     Returns:
#         composites.Pipeline: nodes in a path through 'tree'.
        
#     """         
#     visited = visited or []
#     if hasattr(tree, 'is_root') and tree.is_root:
#         visited.append(tree)
#     if hasattr(tree, 'children') and tree.children:
#         visited.extend(tree.children)
#         for child in tree.children:
#             visited.extend(breadth_first_search(tree = child, visited = visited))
#     return visited
                
                     
def depth_first_search(
    tree: Tree, 
    visited: Optional[list[Tree]] = None) -> composites.Pipeline:
    """Returns a depth first search path through 'tree'.

    Args:
        tree (Tree): tree to search.
        visited (Optional[list[Tree]]): list of visited nodes. Defaults to None.

    Returns:
        composites.Pipeline: nodes in a path through 'tree'.
        
    """  
    visited = visited or []
    visited.append(tree)
    if hasattr(tree, 'children') and tree.children:
        for child in tree.children:
            visited.extend(depth_first_search(tree = child, visited = visited))
    return visited




"""
graph: lightweight graph data structures
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2021, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    System (Graph): a lightweight directed acyclic graph (DAG). Internally, the 
        graph is stored as an adjacency list. As a result, it should primarily 
        be used for workflows or other uses that do require large graphs.
        
To Do:
    Complete Network which will use an adjacency matrix for internal storage.
    
"""
from __future__ import annotations
import abc
import collections
from collections.abc import (
    Collection, Hashable, MutableMapping, MutableSequence, Sequence, Set)
import copy
import dataclasses
import itertools
from typing import Any, Optional, Type, TYPE_CHECKING, Union

from ..base import mappings
from ..observe import traits
from ..repair import convert
from . import composites
from . import check
from . import transform
from . import hybrid

if TYPE_CHECKING:
    from . import tree
    
 
@dataclasses.dataclass
class Edge(composites.Composite, abc.ABC):
    
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        return check.is_edge(item = instance)

     
@dataclasses.dataclass # type: ignore
class Graph(composites.Composite, abc.ABC):
    """composites class for graph data structures.
    
    Args:
        contents (Collection[Any]): stored collection of nodes and/or edges.
                                      
    """  
    contents: Collection[Any]

    """ Required Subclass Properties """

    @abc.abstractmethod
    def connect(self, start: composites.Node, stop: composites.Node) -> None:
        """Creates a new edge in the stored graph.

        Args:
            start (Node): starting node for the new edge.
            stop (Node): ending node for the new edge.
        
        Raises:
            KeyError: if 'start' or 'stop' are not a nodes in the stored graph.

        """
        pass  
    
    @abc.abstractmethod
    def disconnect(self, start: composites.Node, stop: composites.Node) -> None:
        """Deletes edge from the stored graph.

        Args:
            start (Node): starting node for the edge to delete.
            stop (Node): ending node for the edge to delete.
        
        Raises:
            KeyError: if 'start' or 'stop' are not a nodes in the stored graph.

        """
        pass  
    
    """ Dunder Methods """
    
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        return check.is_graph(item = instance)


@dataclasses.dataclass
class Adjacency(mappings.Dictionary, Graph):
    """composites class for adjacency-list-compositesd graphs.
    
    Args:
        contents (MutableMapping[composites.Node, Set[composites.Node]]): keys are nodes and 
            values are sets of nodes (or hashable representations of nodes). 
            Defaults to a defaultdict that has a set for its value format.
                                      
    """  
    contents: MutableMapping[composites.Node, Set[composites.Node]] = dataclasses.field(
        default_factory = lambda: collections.defaultdict(set))
    
    """ Dunder Methods """
    
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        return check.is_adjacency(item = instance)


@dataclasses.dataclass
class Edges(Graph):
    """composites class for edges-list-compositesd graphs.
    
    Args:
        contents (tuple[tuple[composites.Node, Node], ...]): tuple of tuple of edges. 
            Defaults to an empty tuple.
                                      
    """   
    contents: tuple[tuple[composites.Node, composites.Node], ...] = dataclasses.field(
        default_factory = tuple)
    
    """ Dunder Methods """
           
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        return check.is_edges(item = instance)
    
    
@dataclasses.dataclass
class Matrix(Graph):
    """composites class for adjacency-matrix-compositesd graphs.
    
    Args:
        contents (Sequence[Sequence[int]]): a list of list of integers 
            indicating edges between nodes in the matrix. Defaults to an empty
            list.
        labels (Sequence[composites.Node]): names of nodes in the matrix. Defaults to 
            an empty list.
                                      
    """  
    contents: Sequence[Sequence[int]] = dataclasses.field(
        default_factory = list)
    labels: Sequence[composites.Node] = dataclasses.field(default_factory = list)
    
    """ Dunder Methods """
        
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        return check.is_matrix(item = instance)

    
@dataclasses.dataclass
class System(Adjacency):
    """Directed graph with unweighted edges.
    
    Args:
        contents (MutableMapping[composites.Node, Set[composites.Node]]): keys 
            are nodes and values are sets of nodes (or hashable representations 
            of nodes). Defaults to a defaultdict that has a set for its value 
            format.
                  
    """  
    contents: MutableMapping[composites.Node, Set[composites.Node]] = (
        dataclasses.field(
            default_factory = lambda: collections.defaultdict(set)))
    
    """ Properties """

    @property
    def endpoint(self) -> set[composites.Node]:
        """Returns endpoint nodes in the stored graph in a list."""
        return {k for k in self.contents.keys() if not self.contents[k]}
                    
    @property
    def root(self) -> set[composites.Node]:
        """Returns root nodes in the stored graph in a list."""
        stops = list(itertools.chain.from_iterable(self.contents.values()))
        return {k for k in self.contents.keys() if k not in stops}

    @property
    def adjacency(self) -> Adjacency:
        """Returns the stored graph as an adjacency list."""
        return self.contents

    @property
    def edges(self) -> Edges:
        """Returns the stored graph as an edge list."""
        return transform.adjacency_to_edges(item = self.contents)
    
    @property
    def matrix(self) -> Matrix:
        """Returns the stored graph as an adjacency matrix."""
        return transform.adjacency_to_matrix(item = self.contents)
                      
    @property
    def nodes(self) -> set[composites.Node]:
        """Returns all stored nodes in a list."""
        return set(self.contents.keys())

    @property
    def paths(self) -> composites.Nodes:
        """Returns all paths through the stored as a list of nodes."""
        return self._find_all_paths(starts = self.root, stops = self.endpoint)
    
    @property
    def pipeline(self) -> hybrid.Pipeline:
        """Returns stored graph as a pipeline."""
        pipeline = []
        for pipe in self.pipelines.values():
            pipeline.extend(pipe)
        return hybrid.Pipeline(contents = pipeline)
    
    @property
    def pipelines(self) -> hybrid.Pipelines:
        """Returns stored graph as pipelines."""
        all_paths = self.paths
        instances = [hybrid.Pipeline(contents = p) for p in all_paths]
        pipelines = hybrid.Pipelines()
        for instance in instances:
            pipelines.add(instance, name = 'path')
        return pipelines
            
    @property
    def tree(self) -> tree.Tree:
        """Returns the stored composite object as a tree.Tree."""
        raise NotImplementedError

    """ Class Methods """
 
    @classmethod
    def from_adjacency(cls, item: Adjacency) -> System:
        """Creates a System instance from an adjacency list."""
        return cls(contents = item)
    
    @classmethod
    def from_edges(cls, item: Edges) -> System:
        """Creates a System instance from an edge list."""
        return cls(contents = transform.edges_to_adjacency(item = item))
    
    @classmethod
    def from_matrix(cls, item: Matrix) -> System:
        """Creates a System instance from an adjacency matrix."""
        return cls(contents = transform.matrix_to_adjacency(item = item))
    
    @classmethod
    def from_nodes(cls, item: composites.Nodes) -> System:
        """Creates a System instance from a Nodes."""
        new_contents = transform.pipeline_to_adjacency(item = item)
        return cls(contents = new_contents)

    @classmethod
    def from_pipeline(cls, item: hybrid.Pipeline) -> System:
        """Creates a System instance from a Pipeline."""
        new_contents = transform.pipeline_to_adjacency(item = item)
        return cls(contents = new_contents)
    
    @classmethod
    def from_pipelines(cls, item: hybrid.Pipelines) -> System:
        """Creates a System instance from a Pipeline."""
        new_contents = transform.pipelines_to_adjacency(item = item)
        return cls(contents = new_contents)

    @classmethod
    def from_tree(cls, item: tree.Tree) -> System:
        """Creates a System instance from a Tree."""
        raise NotImplementedError
             
    """ Public Methods """

    def add(
        self, 
        node: composites.Node,
        ancestors: composites.Nodes = None,
        descendants: composites.Nodes = None) -> None:
        """Adds 'node' to the stored graph.
        
        Args:
            node (composites.Node): a node to add to the stored graph.
            ancestors (composites.Nodes): node(s) from which 'node' should be 
                connected.
            descendants (composites.Nodes): node(s) to which 'node' should be 
                connected.

        Raises:
            KeyError: if some nodes in 'descendants' or 'ancestors' are not in 
                the stored graph.
                
        """
        if descendants is None:
            self.contents[node] = set()
        # elif utilities.is_property(item = descendants, instance = self):
        #     self.contents = set(getattr(self, descendants))
        else:
            descendants = list(convert.iterify(item = descendants))
            descendants = [traits.get_name(item = n) for n in descendants]
            missing = [n for n in descendants if n not in self.contents]
            if missing:
                raise KeyError(
                    f'descendants {str(missing)} are not in '
                    f'{self.__class__.__name__}')
            else:
                self.contents[node] = set(descendants)
        if ancestors is not None:  
            # if utilities.is_property(item = ancestors, instance = self):
            #     start = list(getattr(self, ancestors))
            # else:
            ancestors = list(convert.iterify(item = ancestors))
            missing = [n for n in ancestors if n not in self.contents]
            if missing:
                raise KeyError(
                    f'ancestors {str(missing)} are not in '
                    f'{self.__class__.__name__}')
            for start in ancestors:
                if node not in self[start]:
                    self.connect(start = start, stop = node)                 
        return 

    def append(self, item: composites.Composite) -> None:
        """Appends 'item' to the endpoints of the stored graph.

        Appending creates an edge between every endpoint of this instance's
        stored graph and the every root of 'item'.

        Args:
            item (composites.Composite): another Graph, 
                an adjacency list, an edge list, an adjacency matrix, or one or
                more nodes.
            
        Raises:
            TypeError: if 'item' is neither a Graph, Adjacency, Edges, Matrix,
                or composites.Nodes type.
                
        """
        if isinstance(item, composites.Composite):
            current_endpoints = list(self.endpoint)
            new_graph = self.create(item = item)
            self.merge(item = new_graph)
            for endpoint in current_endpoints:
                for root in new_graph.root:
                    self.connect(start = endpoint, stop = root)
        else:
            raise TypeError('item must be a Node, Nodes, or Composite type')
        return
  
    def connect(self, start: composites.Node, stop: composites.Node) -> None:
        """Adds an edge from 'start' to 'stop'.

        Args:
            start (composites.Node): name of node for edge to start.
            stop (composites.Node): name of node for edge to stop.
            
        Raises:
            ValueError: if 'start' is the same as 'stop'.
            
        """
        if start == stop:
            raise ValueError(
                'The start of an edge cannot be the same as the '
                'stop in a System because it is acyclic')
        elif start not in self:
            self.add(node = start)
        elif stop not in self:
            self.add(node = stop)
        if stop not in self.contents[start]:
            self.contents[start].add(traits.get_name(item = stop))
        return

    def delete(self, node: composites.Node) -> None:
        """Deletes node from graph.
        
        Args:
            node (composites.Node): node to delete from 'contents'.
        
        Raises:
            KeyError: if 'node' is not in 'contents'.
            
        """
        try:
            del self.contents[node]
        except KeyError:
            raise KeyError(f'{node} does not exist in the graph')
        self.contents = {k: v.discard(node) for k, v in self.contents.items()}
        return

    def disconnect(self, start: composites.Node, stop: composites.Node) -> None:
        """Deletes edge from graph.

        Args:
            start (composites.Node): starting node for the edge to delete.
            stop (composites.Node): ending node for the edge to delete.
        
        Raises:
            KeyError: if 'start' is not a node in the stored graph..

        """
        try:
            self.contents[start].discard(stop)
        except KeyError:
            raise KeyError(f'{start} does not exist in the graph')
        return

    def merge(self, item: composites.Composite) -> None:
        """Adds 'item' to this Graph.

        This method is roughly equivalent to a dict.update, just adding the
        new keys and values to the existing graph. It transforms 'item' to an 
        adjacency list that is then added to the existing 'contents'.
        
        Args:
            item (composites.Composite): another Graph, an adjacency 
                list, an edge list, an adjacency matrix, or one or more nodes.
            
        Raises:
            TypeError: if 'item' is neither a System, Adjacency, 
                Edges, Matrix, or composites.Nodes type.
            
        """
        if isinstance(item, System):
            adjacency = item.adjacency
        elif isinstance(item, Adjacency):
            adjacency = item
        elif isinstance(item, Edges):
            adjacency = transform.edges_to_adjacency(item = item)
        elif isinstance(item, Matrix):
            adjacency = transform.matrix_to_adjacency(item = item)
        elif isinstance(item, (list, tuple, set)):
            adjacency = transform.pipeline_to_adjacency(item = item)
        elif isinstance(item, composites.Node):
            adjacency = {item: set()}
        else:
            raise TypeError('item must be a Node, Nodes, or Composite type')
        self.contents.update(adjacency)
        return

    def prepend(self, item: composites.Composite) -> None:
        """Prepends 'item' to the roots of the stored graph.

        Prepending creates an edge between every endpoint of 'item' and every
        root of this instance;s stored graph.

        Args:
            item (composites.Composite): another Graph, an adjacency list, an 
                edge list, an adjacency matrix, or one or more nodes.
            
        Raises:
            TypeError: if 'item' is neither a System, Adjacency, Edges, Matrix, 
                or composites.Nodes type.
                
        """
        if isinstance(item, composites.Composite):
            current_roots = list(self.root)
            new_graph = self.create(item = item)
            self.merge(item = new_graph)
            for root in current_roots:
                for endpoint in new_graph.endpoints:
                    self.connect(start = endpoint, stop = root)
        else:
            raise TypeError(
                'item must be a System, Adjacency, Edges, Matrix, hybrid.Pipeline, '
                'hybrid.Pipelines, or Node type')
        return
      
    def subset(
        self, 
        include: Union[Any, Sequence[Any]] = None,
        exclude: Union[Any, Sequence[Any]] = None) -> System:
        """Returns a new System without a subset of 'contents'.
        
        All edges will be removed that include any nodes that are not part of
        the new subgraph.
        
        Any extra attributes that are part of a System (or a subclass) will be
        maintained in the returned subgraph.

        Args:
            include (Union[Any, Sequence[Any]]): nodes which should be included
                with any applicable edges in the new subgraph.
            exclude (Union[Any, Sequence[Any]]): nodes which should not be 
                included with any applicable edges in the new subgraph.

        Returns:
           System: with only key/value pairs with keys not in 'subset'.

        """
        if include is None and exclude is None:
            raise ValueError('Either include or exclude must not be None')
        else:
            if include:
                excludables = [k for k in self.contents if k not in include]
            else:
                excludables = []
            excludables.extend([i for i in self.contents if i in exclude])
            new_graph = copy.deepcopy(self)
            for node in convert.iterify(item = excludables):
                new_graph.delete(node = node)
        return new_graph
    
    def walk(
        self, 
        start: composites.Node,
        stop: composites.Node, 
        path: Optional[hybrid.Pipeline] = None) -> hybrid.Pipeline:
        """Returns all paths in graph from 'start' to 'stop'.

        The code here is adapted from: https://www.python.org/doc/essays/graphs/
        
        Args:
            start (composites.Node): node to start paths from.
            stop (composites.Node): node to stop paths.
            path (hybrid.Pipeline): a path from 'start' to 'stop'. Defaults 
                to an empty list. 

        Returns:
            hybrid.Pipeline: a list of possible paths (each path is a list 
                nodes) from 'start' to 'stop'.
            
        """
        if path is None:
            path = []
        path = path + [start]
        if start == stop:
            return [path]
        if start not in self.contents:
            return []
        paths = []
        for node in self.contents[start]:
            if node not in path:
                new_paths = self.walk(
                    start = node, 
                    stop = stop, 
                    path = path)
                for new_path in new_paths:
                    paths.append(new_path)
        return paths

    """ Private Methods """

    def _find_all_paths(
        self, 
        starts: composites.Nodes, 
        stops: composites.Nodes) -> hybrid.Pipeline:
        """Returns all paths between 'starts' and 'stops'.

        Args:
            start (composites.Nodes): starting point(s) for paths through the 
                System.
            ends (composites.Nodes): ending point(s) for paths through the 
                System.

        Returns:
            hybrid.Pipeline: list of all paths through the System from all 
                'starts' to all 'ends'.
            
        """
        all_paths = []
        for start in convert.iterify(item = starts):
            for end in convert.iterify(item = stops):
                paths = self.walk(start = start, stop = end)
                if paths:
                    if all(isinstance(path, composites.Node) for path in paths):
                        all_paths.append(paths)
                    else:
                        all_paths.extend(paths)
        return all_paths


# @dataclasses.dataclass
# class Network(Graph):
#     """composites class for undirected graphs with unweighted edges.
    
#     Graph stores a directed acyclic graph (DAG) as an adjacency list. Despite 
#     being called an adjacency "list," the typical and most efficient way to 
#     store one is using a python dict. a piles Graph inherits from a Dictionary 
#     in order to allow use of its extra functionality over a plain dict.
    
#     Graph supports '+' and '+=' to be used to join two piles Graph instances. A
#     properly formatted adjacency list could also be the added object.
    
#     Graph internally supports autovivification where a list is created as a 
#     value for a missing key. This means that a Graph need not inherit from 
#     defaultdict.
    
#     Args:
#         contents (Adjacency): an adjacency list where the keys are nodes and the 
#             values are nodes which the key is connected to. Defaults to an empty 
#             dict.
                  
#     """  
#     contents: Matrix = dataclasses.field(default_factory = dict)
    
#     """ Properties """

#     @property
#     def adjacency(self) -> Adjacency:
#         """Returns the stored graph as an adjacency list."""
#         return matrix_to_adjacency(item = self.contents)

#     @property
#     def breadths(self) -> hybrid.Pipeline:
#         """Returns all paths through the Graph using breadth-first search.
        
#         Returns:
#             hybrid.Pipeline: returns all paths from 'roots' to 'endpoints' in a list 
#                 of lists of nodes.
                
#         """
#         return self._find_all_paths(
#             starts = self.root, 
#             ends = self.endpoint,
#             depth_first = False)

#     @property
#     def depths(self) -> hybrid.Pipeline:
#         """Returns all paths through the Graph using depth-first search.
        
#         Returns:
#             hybrid.Pipeline: returns all paths from 'roots' to 'endpoints' in a list 
#                 of lists of nodes.
                
#         """
#         return self._find_all_paths(starts = self.root, 
#                                     ends = self.endpoint,
#                                     depth_first = True)
     
#     @property
#     def edges(self) -> Edges:
#         """Returns the stored graph as an edge list."""
#         return adjacency_to_edges(item = self.contents)

#     @property
#     def endpoints(self) -> list[composites.Node]:
#         """Returns a list of endpoint nodes in the stored graph.."""
#         return [k for k in self.contents.keys() if not self.contents[k]]

#     @property
#     def matrix(self) -> Matrix:
#         """Returns the stored graph as an adjacency matrix."""
#         return adjacency_to_matrix(item = self.contents)
                      
#     @property
#     def nodes(self) -> dict[str, composites.Node]:
#         """Returns a dict of node names as keys and nodes as values.
        
#         Because Graph allows various composites.Node objects to be used as keys,
#         including the composites.Nodes class, there isn't an obvious way to access already
#         stored nodes. This property creates a new dict with str keys derived
#         from the nodes (looking first for a 'name' attribute) so that a user
#         can access a node. 
        
#         This property is not needed if the stored nodes are all strings.
        
#         Returns:
#             Dict[str, composites.Node]: keys are the name or has of nodes and the 
#                 values are the nodes themselves.
            
#         """
#         return {self.traits.get_name(item = n): n for n in self.contents.keys()}
  
#     @property
#     def roots(self) -> list[composites.Node]:
#         """Returns root nodes in the stored graph..

#         Returns:
#             list[composites.Node]: root nodes.
            
#         """
#         stops = list(itertools.chain.from_iterable(self.contents.values()))
#         return [k for k in self.contents.keys() if k not in stops]
    
#     """ Class Methods """
    
#     @classmethod
#     def create(cls, item: Union[Adjacency, Edges, Matrix]) -> Graph:
#         """Creates an instance of a Graph from 'item'.
        
#         Args:
#             item (Union[Adjacency, Edges, Matrix]): an adjacency list, 
#                 adjacency matrix, or edge list which can used to create the
#                 stored graph.
                
#         Returns:
#             Graph: a Graph instance created compositesd on 'item'.
                
#         """
#         if is_adjacency_list(item = item):
#             return cls.from_adjacency(adjacency = item)
#         elif is_adjacency_matrix(item = item):
#             return cls.from_matrix(matrix = item)
#         elif is_edge_list(item = item):
#             return cls.from_adjacency(edges = item)
#         else:
#             raise TypeError(
#                 f'create requires item to be an adjacency list, adjacency '
#                 f'matrix, or edge list')
           
#     @classmethod
#     def from_adjacency(cls, adjacency: Adjacency) -> Graph:
#         """Creates a Graph instance from an adjacency list.
        
#         'adjacency' should be formatted with nodes as keys and values as lists
#         of names of nodes to which the node in the key is connected.

#         Args:
#             adjacency (Adjacency): adjacency list used to 
#                 create a Graph instance.

#         Returns:
#             Graph: a Graph instance created compositesd on 'adjacent'.
              
#         """
#         return cls(contents = adjacency)
    
#     @classmethod
#     def from_edges(cls, edges: Edges) -> Graph:
#         """Creates a Graph instance from an edge list.

#         'edges' should be a list of tuples, where the first item in the tuple
#         is the node and the second item is the node (or name of node) to which
#         the first item is connected.
        
#         Args:
#             edges (Edges): Edge list used to create a Graph 
#                 instance.
                
#         Returns:
#             Graph: a Graph instance created compositesd on 'edges'.

#         """
#         return cls(contents = edges_to_adjacency(item = edges))
    
#     @classmethod
#     def from_matrix(cls, matrix: Matrix) -> Graph:
#         """Creates a Graph instance from an adjacency matrix.

#         Args:
#             matrix (Matrix): adjacency matrix used to create a Graph instance. 
#                 The values in the matrix should be 1 (indicating an edge) and 0 
#                 (indicating no edge).
 
#         Returns:
#             Graph: a Graph instance created compositesd on 'matrix'.
                        
#         """
#         return cls(contents = matrix_to_adjacency(item = matrix))
    
#     @classmethod
#     def from_pipeline(cls, pipeline: hybrid.Pipeline) -> Graph:
#         """Creates a Graph instance from a Pipeline.

#         Args:
#             pipeline (hybrid.Pipeline): serial pipeline used to create a Graph
#                 instance.
 
#         Returns:
#             Graph: a Graph instance created compositesd on 'pipeline'.
                        
#         """
#         return cls(contents = pipeline_to_adjacency(item = pipeline))
       
#     """ Public Methods """
    
#     def add(self, 
#             node: composites.Node,
#             ancestors: composites.Nodes = None,
#             descendants: composites.Nodes = None) -> None:
#         """Adds 'node' to 'contents' with no corresponding edges.
        
#         Args:
#             node (composites.Node): a node to add to the stored graph.
#             ancestors (composites.Nodes): node(s) from which node should be connected.
#             descendants (composites.Nodes): node(s) to which node should be connected.

#         """
#         if descendants is None:
#             self.contents[node] = []
#         elif descendants in self:
#             self.contents[node] = convert.iterify(item = descendants)
#         else:
#             missing = [n for n in descendants if n not in self.contents]
#             raise KeyError(f'descendants {missing} are not in the stored graph.')
#         if ancestors is not None:  
#             if (isinstance(ancestors, composites.Node) and ancestors in self
#                     or (isinstance(ancestors, (list, tuple, set)) 
#                         and all(isinstance(n, composites.Node) for n in ancestors)
#                         and all(n in self.contents for n in ancestors))):
#                 start = ancestors
#             elif (hasattr(self.__class__, ancestors) 
#                     and isinstance(getattr(type(self), ancestors), property)):
#                 start = getattr(self, ancestors)
#             else:
#                 missing = [n for n in ancestors if n not in self.contents]
#                 raise KeyError(f'ancestors {missing} are not in the stored graph.')
#             for starting in convert.iterify(item = start):
#                 if node not in [starting]:
#                     self.connect(start = starting, stop = node)                 
#         return 

#     def append(self, 
#                item: Union[Graph, Adjacency, Edges, Matrix, composites.Nodes]) -> None:
#         """Adds 'item' to this Graph.

#         Combining creates an edge between every endpoint of this instance's
#         Graph and the every root of 'item'.

#         Args:
#             item (Union[Graph, Adjacency, Edges, Matrix, composites.Nodes]): another 
#                 Graph to join with this one, an adjacency list, an edge list, an
#                 adjacency matrix, or composites.Nodes.
            
#         Raises:
#             TypeError: if 'item' is neither a Graph, Adjacency, Edges, Matrix,
#                 or composites.Nodes type.
            
#         """
#         if isinstance(item, Graph):
#             if self.contents:
#                 current_endpoints = self.endpoint
#                 self.contents.update(item.contents)
#                 for endpoint in current_endpoints:
#                     for root in item.root:
#                         self.connect(start = endpoint, stop = root)
#             else:
#                 self.contents = item.contents
#         elif isinstance(item, Adjacency):
#             self.append(item = self.from_adjacency(adjacecny = item))
#         elif isinstance(item, Edges):
#             self.append(item = self.from_edges(edges = item))
#         elif isinstance(item, Matrix):
#             self.append(item = self.from_matrix(matrix = item))
#         elif isinstance(item, composites.Nodes):
#             if isinstance(item, (list, tuple, set)):
#                 new_graph = Graph()
#                 edges = more_itertools.windowed(item, 2)
#                 for edge_pair in edges:
#                     new_graph.add(node = edge_pair[0], descendants = edge_pair[1])
#                 self.append(item = new_graph)
#             else:
#                 self.add(node = item)
#         else:
#             raise TypeError(
#                 'item must be a Graph, Adjacency, Edges, Matrix, or composites.Nodes '
#                 'type')
#         return
  
#     def connect(self, start: composites.Node, stop: composites.Node) -> None:
#         """Adds an edge from 'start' to 'stop'.

#         Args:
#             start (composites.Node): name of node for edge to start.
#             stop (composites.Node): name of node for edge to stop.
            
#         Raises:
#             ValueError: if 'start' is the same as 'stop'.
            
#         """
#         if start == stop:
#             raise ValueError(
#                 'The start of an edge cannot be the same as the stop')
#         else:
#             if stop not in self.contents:
#                 self.add(node = stop)
#             if start not in self.contents:
#                 self.add(node = start)
#             if stop not in self.contents[start]:
#                 self.contents[start].append(self.traits.get_name(item = stop))
#         return

#     def delete(self, node: composites.Node) -> None:
#         """Deletes node from graph.
        
#         Args:
#             node (composites.Node): node to delete from 'contents'.
        
#         Raises:
#             KeyError: if 'node' is not in 'contents'.
            
#         """
#         try:
#             del self.contents[node]
#         except KeyError:
#             raise KeyError(f'{node} does not exist in the graph')
#         self.contents = {
#             k: v.remove(node) for k, v in self.contents.items() if node in v}
#         return

#     def disconnect(self, start: composites.Node, stop: composites.Node) -> None:
#         """Deletes edge from graph.

#         Args:
#             start (composites.Node): starting node for the edge to delete.
#             stop (composites.Node): ending node for the edge to delete.
        
#         Raises:
#             KeyError: if 'start' is not a node in the stored graph..
#             ValueError: if 'stop' does not have an edge with 'start'.

#         """
#         try:
#             self.contents[start].remove(stop)
#         except KeyError:
#             raise KeyError(f'{start} does not exist in the graph')
#         except ValueError:
#             raise ValueError(f'{stop} is not connected to {start}')
#         return

#     def merge(self, item: Union[Graph, Adjacency, Edges, Matrix]) -> None:
#         """Adds 'item' to this Graph.

#         This method is roughly equivalent to a dict.update, just adding the
#         new keys and values to the existing graph. It transforms the supported
#         formats to an adjacency list that is then added to the existing 
#         'contents'.
        
#         Args:
#             item (Union[Graph, Adjacency, Edges, Matrix]): another Graph to 
#                 add to this one, an adjacency list, an edge list, or an
#                 adjacency matrix.
            
#         Raises:
#             TypeError: if 'item' is neither a Graph, Adjacency, Edges, or 
#                 Matrix type.
            
#         """
#         if isinstance(item, Graph):
#             item = item.contents
#         elif isinstance(item, Adjacency):
#             pass
#         elif isinstance(item, Edges):
#             item = self.from_edges(edges = item).contents
#         elif isinstance(item, Matrix):
#             item = self.from_matrix(matrix = item).contents
#         else:
#             raise TypeError(
#                 'item must be a Graph, Adjacency, Edges, or Matrix type to '
#                 'update')
#         self.contents.update(item)
#         return
  
#     def subgraph(self, 
#                  include: Union[Any, Sequence[Any]] = None,
#                  exclude: Union[Any, Sequence[Any]] = None) -> Graph:
#         """Returns a new Graph without a subset of 'contents'.
        
#         All edges will be removed that include any nodes that are not part of
#         the new subgraph.
        
#         Any extra attributes that are part of a Graph (or a subclass) will be
#         maintained in the returned subgraph.

#         Args:
#             include (Union[Any, Sequence[Any]]): nodes which should be included
#                 with any applicable edges in the new subgraph.
#             exclude (Union[Any, Sequence[Any]]): nodes which should not be 
#                 included with any applicable edges in the new subgraph.

#         Returns:
#             Graph: with only key/value pairs with keys not in 'subset'.

#         """
#         if include is None and exclude is None:
#             raise ValueError('Either include or exclude must not be None')
#         else:
#             if include:
#                 excludables = [k for k in self.contents if k not in include]
#             else:
#                 excludables = []
#             excludables.extend([i for i in self.contents if i not in exclude])
#             new_graph = copy.deepcopy(self)
#             for node in convert.iterify(item = excludables):
#                 new_graph.delete(node = node)
#         return new_graph

#     def walk(self, 
#              start: composites.Node, 
#              stop: composites.Node, 
#              path: hybrid.Pipeline = None,
#              depth_first: bool = True) -> hybrid.Pipeline:
#         """Returns all paths in graph from 'start' to 'stop'.

#         The code here is adapted from: https://www.python.org/doc/essays/graphs/
        
#         Args:
#             start (composites.Node): node to start paths from.
#             stop (composites.Node): node to stop paths.
#             path (hybrid.Pipeline): a path from 'start' to 'stop'. Defaults to an 
#                 empty list. 

#         Returns:
#             hybrid.Pipeline: a list of possible paths (each path is a list 
#                 nodes) from 'start' to 'stop'.
            
#         """
#         if path is None:
#             path = []
#         path = path + [start]
#         if start == stop:
#             return [path]
#         if start not in self.contents:
#             return []
#         if depth_first:
#             method = self._depth_first_search
#         else:
#             method = self._breadth_first_search
#         paths = []
#         for node in self.contents[start]:
#             if node not in path:
#                 new_paths = self.walk(
#                     start = node, 
#                     stop = stop, 
#                     path = path,
#                     depth_first = depth_first)
#                 for new_path in new_paths:
#                     paths.append(new_path)
#         return paths

#     def _all_paths_bfs(self, start, stop):
#         """

#         """
#         if start == stop:
#             return [start]
#         visited = {start}
#         queue = collections.deque([(start, [])])
#         while queue:
#             current, path = queue.popleft()
#             visited.add(current)
#             for connected in self[current]:
#                 if connected == stop:
#                     return path + [current, connected]
#                 if connected in visited:
#                     continue
#                 queue.append((connected, path + [current]))
#                 visited.add(connected)   
#         return []

#     def _breadth_first_search(self, node: composites.Node) -> hybrid.Pipeline:
#         """Returns a breadth first search path through the Graph.

#         Args:
#             node (composites.Node): node to start the search from.

#         Returns:
#             hybrid.Pipeline: nodes in a path through the Graph.
            
#         """        
#         visited = set()
#         queue = [node]
#         while queue:
#             vertex = queue.pop(0)
#             if composites. not in visited:
#                 visited.add(vertex)
#                 queue.extend(set(self[vertex]) - visited)
#         return list(visited)
       
#     def _depth_first_search(self, 
#         node: composites.Node, 
#         visited: list[composites.Node]) -> hybrid.Pipeline:
#         """Returns a depth first search path through the Graph.

#         Args:
#             node (composites.Node): node to start the search from.
#             visited (list[composites.Node]): list of visited nodes.

#         Returns:
#             hybrid.Pipeline: nodes in a path through the Graph.
            
#         """  
#         if node not in visited:
#             visited.append(node)
#             for edge in self[node]:
#                 self._depth_first_search(node = edge, visited = visited)
#         return visited
  
#     def _find_all_paths(self, 
#         starts: Union[composites.Node, Sequence[composites.Node]],
#         stops: Union[composites.Node, Sequence[composites.Node]],
#         depth_first: bool = True) -> hybrid.Pipeline:
#         """[summary]

#         Args:
#             start (Union[composites.Node, Sequence[composites.Node]]): starting points for 
#                 paths through the Graph.
#             ends (Union[composites.Node, Sequence[composites.Node]]): endpoints for paths 
#                 through the Graph.

#         Returns:
#             hybrid.Pipeline: list of all paths through the Graph from all
#                 'starts' to all 'ends'.
            
#         """
#         all_paths = []
#         for start in convert.iterify(item = starts):
#             for end in convert.iterify(item = stops):
#                 paths = self.walk(
#                     start = start, 
#                     stop = end,
#                     depth_first = depth_first)
#                 if paths:
#                     if all(isinstance(path, composites.Node) for path in paths):
#                         all_paths.append(paths)
#                     else:
#                         all_paths.extend(paths)
#         return all_paths
            
#     """ Dunder Methods """

#     def __add__(self, other: Graph) -> None:
#         """Adds 'other' Graph to this Graph.

#         Adding another graph uses the 'merge' method. Read that method's 
#         docstring for further details about how the graphs are added 
#         together.
        
#         Args:
#             other (Graph): a second Graph to join with this one.
            
#         """
#         self.merge(graph = other)        
#         return

#     def __iadd__(self, other: Graph) -> None:
#         """Adds 'other' Graph to this Graph.

#         Adding another graph uses the 'merge' method. Read that method's 
#         docstring for further details about how the graphs are added 
#         together.
        
#         Args:
#             other (Graph): a second Graph to join with this one.
            
#         """
#         self.merge(graph = other)        
#         return

#     def __contains__(self, nodes: composites.Nodes) -> bool:
#         """[summary]

#         Args:
#             nodes (composites.Nodes): [description]

#         Returns:
#             bool: [description]
            
#         """
#         if isinstance(nodes, (list, tuple, set)):
#             return all(n in self.contents for n in nodes)
#         elif isinstance(nodes, composites.Node):
#             return nodes in self.contents
#         else:
#             return False   
        
#     def __getitem__(self, key: composites.Node) -> Any:
#         """Returns value for 'key' in 'contents'.

#         Args:
#             key (composites.Node): key in 'contents' for which a value is sought.

#         Returns:
#             Any: value stored in 'contents'.

#         """
#         return self.contents[key]

#     def __setitem__(self, key: composites.Node, value: Any) -> None:
#         """sets 'key' in 'contents' to 'value'.

#         Args:
#             key (composites.Node): key to set in 'contents'.
#             value (Any): value to be paired with 'key' in 'contents'.

#         """
#         self.contents[key] = value
#         return

#     def __delitem__(self, key: composites.Node) -> None:
#         """Deletes 'key' in 'contents'.

#         Args:
#             key (composites.Node): key in 'contents' to delete the key/value pair.

#         """
#         del self.contents[key]
#         return

#     def __missing__(self) -> list:
#         """Returns an empty list when a key doesn't exist.

#         Returns:
#             list: an empty list.

#         """
#         return []
    
#     def __str__(self) -> str:
#         """Returns prettier summary of the Graph.

#         Returns:
#             str: a formatted str of class information and the contained 
#                 adjacency list.
            
#         """
#         new_line = '\n'
#         tab = '    '
#         summary = [f'{new_line}piles {self.__class__.__name__}']
#         summary.append('adjacency list:')
#         for node, edges in self.contents.items():
#             summary.append(f'{tab}{node}: {str(edges)}')
#         return new_line.join(summary) 




"""
core: classes for easy introspection
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2021, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:   
    Inspector (object):
    ClassInspector (Inspector):
    InstanceInspector (Inspector):
    ModuleInspector (Inspector):
    PackageInspector (Inspector):
  
ToDo:


"""
from __future__ import annotations
import dataclasses
import inspect
import pathlib
import types
from typing import Any, Optional, Type, Union

from . import module
from . import package
from . import traits


@dataclasses.dataclass
class Inspector(object):
    """Inspector factory which returns the appropraite Inspector subclass.
    
    Args:
        item (Any): unknown item to examine.
                        
    """
    item: Any
    
    """ Initialization Methods """
    
    def __new__(cls, item: Any, *args: Any, **kwargs: Any) -> None:
        """Returns Inspector subclass based on type of 'item'."""
        if isinstance(item, types.ModuleType):
            return ModuleInspector(module = item, *args, **kwargs)
        elif isinstance(item, (pathlib.Path, str)):
            return PackageInspector(folder = item, *args, **kwargs)
        elif inspect.isclass(item):
            return ClassInspector(item = item, *args, **kwargs)
        elif isinstance(item, object):
            return InstanceInspector(item = item, *args, **kwargs)
        else:
            raise TypeError(f'item must be a module, path, class, or object')
    
    """ Properties """
        
    @property
    def attributes(self) -> dict[str, Any]:
        """dict of attribute names and values in 'item'.

        Returns:
            dict[str, Any]: keys are attribute names and values are attribute 
                values.
            
        """
        return traits.get_attributes(
            item = self.item, 
            include_private = self.include_private)
        
    @property
    def contains(self) -> Optional[Union[
        tuple[Type[Any], ...], 
        tuple[tuple[Type[Any], ...], tuple[Type[Any], ...]]]]:
        """Types that 'item' contains.
        
        Returns:
            Optional[Union[tuple[Type[Any], ...], tuple[tuple[Type[Any], ...], 
                tuple[Type[Any], ...]]]]:: returns the types of things contained 
                in 'item'. Returns None if 'item' is not a container.
            
        """
        return traits.has_types(item = self.item)
                
    @property
    def name(self) -> Optional[str]:
        """str name of 'item'.
        
        Returns:
            str: inferred name of the stored 'item'.
            
        """
        return traits.get_name(item = self.item) 
        
    @property
    def type(self) -> Type[Any]:
        """Data type of 'item'.

        Returns:
            Type[Any]: type of the stored 'item'.
            
        """
        return type(self.item)
        
    @property
    def variables(self) -> dict[str, Any]:
        """dict of variable names and variable values in 'item'.
        
        'variables' are all attributes that are neither methods nor properties.

        Returns:
            dict[str, Any]: keys are variable names and values are variable 
                values.
            
        """
        return traits.get_variables(
            item = self.item, 
            include_private = self.include_private)                  


@dataclasses.dataclass
class ClassInspector(Inspector):
    """Inspector for accessing class information from 'item'.
    
    Args:
        item (Type[Any]): class to examine.
        include_private (bool): whether to include items that begin with '_'
            (True) or to exclude them (False). Defauls to True.
                        
    """
    item: Type[Any]
    include_private: bool = True
                       
    """ Properties """
    
    @property
    def annotations(self) -> dict[str, Type[Any]]:
        """dict of parameters and annotated type hints in 'item'.

        Returns:
            dict[str, Type[Any]]: keys are parameter/attribute names and values
                are type hints.
            
        """
        return traits.get_annotations(
            item = self.item, 
            include_private = self.include_private) 
        
    @property
    def attributes(self) -> list[str]:
        """Attribute names in 'item'.

        Returns:
            list[str]: names of attributes.
            
        """
        return traits.name_attributes(
            item = self.item, 
            include_private = self.include_private)

    @property
    def methods(self) -> dict[str, types.MethodType]:
        """dict of method names and methods in 'item'.

        Returns:
            dict[str, types.MethodType]: keys are method names and values are 
                methods.
            
        """
        return traits.get_methods(
            item = self.item, 
            include_private = self.include_private)

    @property
    def parameters(self) -> list[str]:
        """Names of parameters from a dataclass: 'item'.
        
        Returns:
            list[str]: names of parameters for a dataclass.
            
        """
        return traits.name_parameters(item = self.item) 
    
    @property
    def properties(self) -> list[str]:
        """Property names in 'item'.

        Returns:
            list[str]: names of properties.
            
        """
        return traits.get_properties(
            item = self.item, 
            include_private = self.include_private)           
                 
    @property
    def signatures(self) -> dict[str, inspect.Signature]:
        """dict of method names and signatures in 'item'.

        Returns:
            dict[str, inspect.Signature]: keys are method names and values are 
                signatures for those methods.
            
        """
        return traits.get_signatures(
            item = self.item, 
            include_private = self.include_private)  

    @property
    def variables(self) -> list[str]:
        """Variable names in 'item'.
        
        'variables' are all attributes that are neither methods nor properties.

        Returns:
            list[str]: names of variables in 'item'.
            
        """
        return traits.get_variables(
            item = self.item, 
            include_private = self.include_private)      


@dataclasses.dataclass
class InstanceInspector(Inspector):
    """Inspector for accessing instance information from 'item'.
    
    Args:
        item (Type[Any]): instance to examine.
        include_private (bool): whether to include items that begin with '_'
            (True) or to exclude them (False). Defauls to True.
                        
    """
    item: object
    include_private: bool = True
             
    """ Properties """
    
    @property
    def annotations(self) -> dict[str, Type[Any]]:
        """dict of parameters and annotated type hints in 'item'.

        Returns:
            dict[str, Type[Any]]: keys are parameter/attribute names and values
                are type hints.
            
        """
        return traits.get_annotations(
            item = self.item, 
            include_private = self.include_private) 

    @property
    def methods(self) -> dict[str, types.MethodType]:
        """dict of method names and methods in 'item'.

        Returns:
            dict[str, types.MethodType]: keys are method names and values are 
                methods.
            
        """
        return traits.get_methods(
            item = self.item, 
            include_private = self.include_private)

    @property
    def parameters(self) -> list[str]:
        """Names of parameters from a dataclass in 'item'.
        
        Returns:
            list[str]: names of parameters for a dataclass.
            
        """
        return traits.name_parameters(item = self.item) 
    
    @property
    def properties(self) -> dict[str, Any]:
        """dict of property names and property values in 'item'.

        Returns:
            dict[str, Any]: keys are property names and values are property 
                values.
            
        """
        return traits.get_properties(
            item = self.item, 
            include_private = self.include_private)           
                 
    @property
    def signatures(self) -> dict[str, inspect.Signature]:
        """dict of method names and signatures in 'item'.

        Returns:
            dict[str, inspect.Signature]: keys are method names and values are 
                signatures for those methods.
            
        """
        return traits.get_signatures(
            item = self.item, 
            include_private = self.include_private)                
         

@dataclasses.dataclass
class ModuleInspector(Inspector):
    """Inspector for accessing module information from 'item'.
    
    Args:
        item (types.ModuleType): module to inspect.
        include_private (bool): whether to include items that begin with '_'
            (True) or to exclude them (True). Defauls to False.
                        
    """
    item: types.ModuleType
    include_private: bool = True
             
    """ Properties """
        
    @property
    def classes(self) -> dict[str, Type[Any]]:
        """dict of class names and classes in 'item'.

        Returns:
            dict[str, Type[Any]: keys are class names and values are classes.
            
        """
        return module.get_classes(
            item = self.item, 
            include_private = self.include_private)
        
    @property
    def functions(self) -> dict[str, types.FunctionType]:
        """dict of function names and functions in 'item'.

        Returns:
            dict[str, types.FunctionType]: keys are function names and values 
                are functions.
            
        """
        return module.get_functions(
            item = self.item, 
            include_private = self.include_private)
                 
    @property
    def signatures(self) -> dict[str, inspect.Signature]:
        """dict of method names and method signatures in 'item'.

        Returns:
            dict[str, inspect.Signature]: keys are method names and values are 
                signatures for those methods.
            
        """
        return traits.get_signatures(
            item = self.item, 
            include_private = self.include_private)  
          

@dataclasses.dataclass
class PackageInspector(Inspector):
    """Inspector for accessing package information from 'item'.
    
    Attributes:
        item (Union[pathlib.Path, str]): folder for which information should
            be made available.
        include_private (bool): whether to include items that begin with '_'
            (True) or to exclude them (False). Defauls to True.
        include_subfolders (bool): whether to include subitems in the package.
            Defaults to True. 
            
    """
    item: Union[pathlib.Path, str]
    include_private: bool = True
    include_subfolders: bool = True

    """ Initialization Methods """
    
    def __post_init__(self) -> None:
        """Initializes class instance attributes."""
        self.item = convert.pathlibify(item = self.item)
        
    """ Properties """

    @property
    def files(self) -> list[pathlib.Path]:
        """Non-python-module file paths in 'item'.

        Returns:
            list[pathlib.Path]: list of non-python-module file paths.
            
        """
        return package.get_file_paths(
            item = self.item, 
            recursive = self.include_subfolders)

    @property    
    def folders(self) -> list[pathlib.Path]:
        """Folder paths in 'item'.

        Returns:
            list[pathlib.Path]: list of folder paths.
            
        """
        return package.get_folder_paths(
            item = self.item, 
            recursive = self.include_subfolders)

    @property          
    def modules(self) -> dict[str, types.ModuleType]:
        """dict of python module names and modules in 'item'.
        
        Args:
            item (Union[str, pathlib.Path]): path of folder to examine.
            
        Returns:
            dict[str, types.ModuleType]: dict with str key names of python 
                modules and values as the corresponding modules.
            
        """
        return package.get_modules(
            item = self.item, 
            recursive = self.include_subfolders)  
        
    @property          
    def module_paths(self) -> list[pathlib.Path]:
        """Python module file paths in 'item'.

        Returns:
            list[pathlib.Path]: list of python-module file paths.
            
        """
        return package.get_module_paths(
            item = self.item, 
            recursive = self.include_subfolders)  
              
    @property    
    def paths(self) -> list[pathlib.Path]:
        """All paths in 'item'.

        Returns:
            list[pathlib.Path]: list of all paths.
            
        """
        return package.get_paths(
            item = self.item, 
            recursive = self.include_subfolders)



"""
convert: functions that convert types
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2021, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

All tools should follow one of two forms. For conversion of a known type
to another type, the function name should be:
    f'{item type}_to_{output type}'
For a conversion from an unknown type to another type, the function name should
be:
    f'to_{output type}'
     
Contents:
    instancify (Callable): converts a class to an instance or adds kwargs to a
        passed instance as attributes.
    listify (Callable): converts passed item to a list.
    namify (Callable): returns hashable name for passed item.
    numify (Callable): attempts to convert passed item to a numerical type.
    pathlibify (Callable): converts a str to a pathlib object or leaves it as
        a pathlib object.
    tuplify (Callable): converts a passed item to a tuple.
    typify (Callable): converts a str type to other common types, if possible.


ToDo:
    Add more flexible tools.
    
"""
from __future__ import annotations
import ast
from collections.abc import (
    Collection, Hashable, Iterable, Mapping, MutableMapping, MutableSequence, 
    Sequence, Set)
import inspect
import pathlib
from typing import Any, Callable, Optional, Type, Union

from . import modify


""" General Converters """

def instancify(item: Union[Type[Any], object], **kwargs: Any) -> Any:
    """Returns 'item' as an instance with 'kwargs' as parameters/attributes.
    
    If 'item' is already an instance, kwargs are added as attributes to the
    existing 'item'. This will overwrite any existing attributes of the same
    name.

    Args:
        item (Union[Type[Any], object])): class to make an instance out of by 
            passing kwargs or an instance to add kwargs to as attributes.

    Raises:
        TypeError: if 'item' is neither a class nor instance.
        
    Returns:
        object: a class instance with 'kwargs' as attributes or passed as 
            parameters (if 'item' is a class).
        
    """         
    if inspect.isclass(item):
        return item(**kwargs) # type: ignore
    elif isinstance(item, object):
        for key, value in kwargs.items():
            setattr(item, key, value)
        return item
    else:
        raise TypeError('item must be a class or class instance')
                  
def iterify(item: Any) -> Iterable:
    """Returns 'item' as an iterable, but does not iterate str types.
    
    Args:
        item (Any): item to turn into an iterable

    Returns:
        Iterable: of 'item'. A str type will be stored as a single item in an
            Iterable wrapper.
        
    """     
    if item is None:
        return iter(())
    elif isinstance(item, (str, bytes)):
        return iter([item])
    else:
        try:
            return iter(item)
        except TypeError:
            return iter((item,))
        
def kwargify(item: Type[Any], args: tuple[Any]) -> dict[Hashable, Any]:
    """Converts args to kwargs.
    
    Args:
        args (tuple): arguments without keywords passed to 'item'.
        item (Type): the item with annotations used to construct kwargs.
    
    Raises:
        ValueError: if there are more args than annotations in 'item'.
        
    Returns
        dict[Hashable, Any]: kwargs based on 'args' and 'item'.
    
    """
    annotations = list(item.__annotations__.keys())
    if len(args) > len(annotations):
        raise ValueError('There are too many args for item')
    else:
        return dict(zip(annotations, args))
    
def listify(item: Any, default: Optional[Any] = None) -> Any:
    """Returns passed item as a list (if not already a list).

    Args:
        item (Any): item to be transformed into a list to allow proper 
            iteration.
        default (Optional[Any]): the default value to return if 'item' is None.
            Unfortunately, to indicate you want None to be the default value,
            you need to put 'None' in quotes. If not passed, 'default' is set to 
            [].

    Returns:
        Any: a passed list, 'item' converted to a list, or the 'default' 
            argument.

    """
    if item is None:
        if default is None:
            return []
        elif default in ['None', 'none']:
            return None
        else:
            return default
    elif isinstance(item, MutableSequence) and not isinstance(item, str):
        return item
    else:
        return [item]
                 
def numify(item: Any, raise_error: bool = False) -> Union[int, float, Any]:
    """Converts 'item' to a numeric type.
    
    If 'item' cannot be converted to a numeric type and 'raise_error' is False, 
        'item' is returned as is.

    Args:
        item (str): item to be converted.
        raise_error (bool): whether to raise a TypeError when conversion to a
            numeric type fails (True) or to simply return 'item' (False). 
            Defaults to False.

    Raises:
        TypeError: if 'item' cannot be converted to a numeric type and 
            'raise_error' is True.
            
    Returns
        Union[int, float, str]: converted to numeric type, if possible.

    """
    try:
        return int(item)
    except ValueError:
        try:
            return float(item)
        except ValueError:
            if raise_error:
                raise TypeError(
                    f'{item} not able to be converted to a numeric type')
            else:
                return item

def pathlibify(item: Union[str, pathlib.Path]) -> pathlib.Path:
    """Converts string 'path' to pathlib.Path object.

    Args:
        item (Union[str, pathlib.Path]): either a string summary of a
            path or a pathlib.Path object.

    Raises:
        TypeError if 'path' is neither a str or pathlib.Path type.

    Returns:
        pathlib.Path object.

    """
    if isinstance(item, str):
        return pathlib.Path(item)
    elif isinstance(item, pathlib.Path):
        return item
    else:
        raise TypeError('item must be str or pathlib.Path type')
           
def stringify(item: Any, default: Optional[Any] = None) -> Any:
    """Converts 'item' to a str from a Sequence.
    
    Args:
        item (Any): item to convert to a str from a list if it is a list.
        default (Any): value to return if 'item' is equivalent to a null
            value when passed. Defaults to None.
    
    Raises:
        TypeError: if 'item' is not a str or list-like object.
        
    Returns:
        Any: str, if item was a list, None or the default value if a null value
            was passed, or the item as it was passed if there previous two 
            conditions don't appply.

    """
    if item is None:
        if default is None:
            return ''
        elif default in ['None', 'none']: 
            return None
        else:
            return default
    elif isinstance(item, str):
        return item
    elif isinstance(item, Sequence):
        return ', '.join(item)
    else:
        raise TypeError('item must be str or a sequence')
    
def tuplify(item: Any, default: Optional[Any] = None) -> Any:
    """Returns passed item as a tuple (if not already a tuple).

    Args:
        item (Any): item to be transformed into a tuple.
        default (Any): the default value to return if 'item' is None.
            Unfortunately, to indicate you want None to be the default value,
            you need to put 'None' in quotes. If not passed, 'default'
            is set to ().

    Returns:
        tuple[Any]: a passed tuple, 'item' converted to a tuple, or 
            'default'.

    """
    if item is None:
        if default is None:
            return tuple()
        elif default in ['None', 'none']:
            return None
        else:
            return default
    elif isinstance(item, tuple):
        return item
    elif isinstance(item, Iterable):
        return tuple(item)
    else:
        return tuple([item])
        
def typify(item: str) -> Union[Sequence[Any], int, float, bool, str]:
    """Converts stings to appropriate, supported datatypes.

    The method converts strings to list (if ', ' is present), int, float,
    or bool datatypes based upon the content of the string. If no
    alternative datatype is found, the item is returned in its original
    form.

    Args:
        item (str): string to be converted to appropriate datatype.

    Returns:
        item (str, list, int, float, or bool): converted item.

    """
    if not isinstance(item, str):
        return item
    else:
        try:
            return int(item)
        except ValueError:
            try:
                return float(item)
            except ValueError:
                if item.lower() in ['true', 'yes']:
                    return True
                elif item.lower() in ['false', 'no']:
                    return False
                elif ', ' in item:
                    item = item.split(', ')
                    return [typify(i) for i in item]
                else:
                    return item

                         
""" Specific Converters """

# @amos.dynamic.dispatcher   
def to_dict(item: Any) -> MutableMapping[Hashable, Any]:
    """Converts 'item' to a MutableMapping.
    
    Args:
        item (Any): item to convert to a MutableMapping.

    Raises:
        TypeError: if 'item' is a type that is not registered.

    Returns:
        MutableMapping: derived from 'item'.

    """
    if isinstance(item, MutableMapping):
        return item
    else:
        raise TypeError(
            f'item cannot be converted because it is an unsupported type: '
            f'{type(item).__name__}')

# @amos.dynamic.dispatcher   
def to_index(item: Any) -> Hashable:
    """Converts 'item' to an Hashable.
    
    Args:
        item (Any): item to convert to a Hashable.

    Raises:
        TypeError: if 'item' is a type that is not registered.

    Returns:
        Hashable: derived from 'item'.

    """
    if isinstance(item, Hashable):
        return item
    else:
        try:
            return hash(item)
        except TypeError:
            try:
                return str(item)
            except TypeError:
                try:
                    return modify.snakify(item.__name__)
                except AttributeError:
                    return modify.snakify(item.__class__.__name__)
                except AttributeError:
                    raise TypeError(f'item cannot be converted because it is ' 
                                    f'an unsupported type: '
                                    f'{type(item).__name__}')

# @to_index.register # type: ignore
def str_to_index(item: str) -> Hashable:
    """[summary]

    Args:
        item (str): [description]

    Returns:
        Hashable: [description]
    """    
    """Converts a str to an Hashable."""
    return item

# @amos.dynamic.dispatcher   
def to_int(item: Any) -> int:
    """Converts 'item' to a pathlib.Path.
    
    Args:
        item (Any): item to convert to a int.

    Raises:
        TypeError: if 'item' is a type that is not registered.

    Returns:
        int: derived from 'item'.

    """
    if isinstance(item, int):
        return item
    else:
        raise TypeError(f'item cannot be converted because it is an '
                        f'unsupported type: {type(item).__name__}')

# @to_int.register # type: ignore
def str_to_int(item: str) -> int:
    """[summary]

    Args:
        item (str): [description]

    Returns:
        int: [description]
    """    
    """Converts a str to an int."""
    return int(item)

# @to_int.register # type: ignore
def float_to_int(item: float) -> int:
    """[summary]

    Args:
        item (float): [description]

    Returns:
        int: [description]
    """    
    """Converts a float to an int."""
    return int(item)

# @amos.dynamic.dispatcher   
def to_list(item: Any) -> list[Any]:
    """Converts 'item' to a list.
    
    Args:
        item (Any): item to convert to a list.

    Raises:
        TypeError: if 'item' is a type that is not registered.

    Returns:
        list[Any]: derived from 'item'.

    """
    if isinstance(item, list[Any]):
        return item
    else:
        raise TypeError(
            f'item cannot be converted because it is an unsupported type: '
            f'{type(item).__name__}')

# @to_list.register # type: ignore
def str_to_list(item: str) -> list[Any]:
    """[summary]

    Args:
        item (str): [description]

    Returns:
        list[Any]: [description]
    """    
    """Converts a str to a list."""
    return ast.literal_eval(item)

# @amos.dynamic.dispatcher   
def to_float(item: Any) -> float:
    """Converts 'item' to a float.
    
    Args:
        item (Any): item to convert to a float.

    Raises:
        TypeError: if 'item' is a type that is not registered.

    Returns:
        float: derived from 'item'.

    """
    if isinstance(item, float):
        return item
    else:
        raise TypeError(
            f'item cannot be converted because it is an unsupported type: '
            f'{type(item).__name__}')

# @to_float.register # type: ignore
def int_to_float(item: int) -> float:
    """[summary]

    Args:
        item (int): [description]

    Returns:
        float: [description]
    """    
    """Converts an int to a float."""
    return float(item)

# @to_float.register # type: ignore
def str_to_float(item: str) -> float:
    """[summary]

    Args:
        item (str): [description]

    Returns:
        float: [description]
    """    
    """Converts a str to a float."""
    return float(item)

# @amos.dynamic.dispatcher   
def to_path(item: Any) -> pathlib.Path:
    """Converts 'item' to a pathlib.Path.
    
    Args:
        item (Any): item to convert to a pathlib.Path.

    Raises:
        TypeError: if 'item' is a type that is not registered.

    Returns:
        pathlib.Path: derived from 'item'.

    """
    if isinstance(item, pathlib.Path):
        return item
    else:
        raise TypeError(
            f'item cannot be converted because it is an unsupported type: '
            f'{type(item).__name__}')

# @to_path.register # type: ignore   
def str_to_path(item: str) -> pathlib.Path:
    """[summary]

    Args:
        item (str): [description]

    Returns:
        pathlib.Path: [description]
    """    
    """Converts a str to a pathlib.Path."""
    return pathlib.pathlib.Path(item)

# @amos.dynamic.dispatcher   
def to_str(item: Any) -> str:
    """Converts 'item' to a str.
    
    Args:
        item (Any): item to convert to a str.

    Raises:
        TypeError: if 'item' is a type that is not registered.

    Returns:
        str: derived from 'item'.

    """
    if isinstance(item, str):
        return item
    else:
        raise TypeError(
            f'item cannot be converted because it is an unsupported type: '
            f'{type(item).__name__}')

# @to_str.register # type: ignore
def int_to_str(item: int) -> str:
    """[summary]

    Args:
        item (int): [description]

    Returns:
        str: [description]
    """    
    """Converts an int to a str."""
    return str(item)

# @to_str.register # type: ignore
def float_to_str(item: float) -> str:
    """[summary]

    Args:
        item (float): [description]

    Returns:
        str: [description]
    """    
    """Converts an float to a str."""
    return str(item)

# @to_str.register # type: ignore
def list_to_str(item: list[Any]) -> str:
    """[summary]

    Args:
        item (list[Any]): [description]

    Returns:
        str: [description]
    """    
    """Converts a list to a str."""
    return ', '.join(item)
   
# @to_str.register 
def none_to_str(item: None) -> str:
    """[summary]

    Args:
        item (None): [description]

    Returns:
        str: [description]
    """    
    """Converts None to a str."""
    return 'None'

# @to_str.register # type: ignore
def path_to_str(item: pathlib.Path) -> str:
    """[summary]

    Args:
        item (pathlib.Path): [description]

    Returns:
        str: [description]
    """    
    """Converts a pathlib.Path to a str."""
    return str(item)



"""
mappings: extensible, flexible, lightweight dict-like classes
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2021, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    Dictionary (Bunch, MutableMapping): bunches's drop-in replacement for a 
        python dict with some added functionality.
    Catalog (Dictionary): wildcard-accepting dict which is primarily intended 
        for storing different options and strategies. It also returns lists of 
        matches if a list of keys is provided.
    Library (MutableMapping): a chained mapping used to registering subclasses
        and instances.
        
ToDo:

       
"""
from __future__ import annotations
from collections.abc import (
    Hashable, Iterator, Mapping, MutableMapping, Sequence)
import copy
import dataclasses
import inspect
from typing import Any, Optional, Type, Union

from ..observe import traits
from ..repair import convert
from . import bunches
                  

_ALL_KEYS: list[Any] = ['all', 'All', ['all'], ['All']]
_DEFAULT_KEYS: list[Any] = [
    'default', 'defaults', 'Default', 'Defaults', ['default'], ['defaults'], 
    ['Default'], ['Defaults']]
_NONE_KEYS: list[Any] = ['none', 'None', ['none'], ['None']]


@dataclasses.dataclass  # type: ignore
class Dictionary(bunches.Bunch, MutableMapping):  # type: ignore
    """Basic bunches dict replacement.
    
    A Dictionary differs from an ordinary python dict in ways inherited from 
    Bunch by requiring 'add' and 'subset' methods, storing data in 'contents', 
    and allowing the '+' operator to join Dictionary instances with other 
    mappings, including Dictionary instances. 
    
    # In addition, it differs in 1 other significant way:
    #     1) When returning 'keys', 'values' and 'items', this class returns them
    #         as tuples instead of KeysView, ValuesView, and ItemsView.
    
    Args:
        contents (MutableMapping[Hashable, Any]): stored dictionary. Defaults 
            to an empty dict.
        default_factory (Optional[Any]): default value to return or default 
            function to call when the 'get' method is used. Defaults to None. 
                          
    """
    contents: MutableMapping[Hashable, Any] = dataclasses.field(
        default_factory = dict)
    default_factory: Optional[Any] = None

    """ Public Methods """
     
    def add(self, item: Mapping[Hashable, Any], **kwargs: Any) -> None:
        """Adds 'item' to the 'contents' attribute.
        
        Args:
            item (Mapping[Hashable, Any]): items to add to 'contents' attribute.
            kwargs: creates a consistent interface even when subclasses have
                additional parameters.
                
        """
        self.contents.update(item, **kwargs)
        return

    @classmethod
    def fromkeys(
        cls, 
        keys: Sequence[Hashable], 
        value: Any, 
        **kwargs: Any) -> Dictionary:
        """Emulates the 'fromkeys' class method from a python dict.

        Args:
            keys (Sequence[Hashable]): items to be keys in a new Dictionary.
            value (Any): the value to use for all values in a new Dictionary.

        Returns:
            Dictionary: formed from 'keys' and 'value'.
            
        """
        return cls(contents = dict.fromkeys(keys, value), **kwargs)
    
    def get(self, key: Hashable, default: Optional[Any] = None) -> Any: # type: ignore
        """Returns value in 'contents' or default options.
        
        Args:
            key (Hashable): key for value in 'contents'.
            default (Optional[Any]): default value to return if 'key' is not 
                found in 'contents'.
        
        Raises:
            KeyError: if 'key' is not in the Dictionary and 'default' and the
                'default_factory' attribute are both None.
        
        Returns:
            Any: value matching key in 'contents' or 'default_factory' value. 
            
        """
        try:
            return self[key]
        except (KeyError, TypeError):
            if default is None:
                if self.default_factory is None:
                    raise KeyError(f'{key} is not in the Dictionary')
                else:
                    try:
                        return self.default_factory()
                    except TypeError:
                        return self.default_factory
            else:
                return default
                
    # def items(self) -> tuple[tuple[Hashable, Any], ...]: # type: ignore
    #     """Emulates python dict 'items' method.
        
    #     Returns:
    #         tuple[tuple[Hashable], Any]: a tuple equivalent to dict.items(). 
            
    #     """
    #     return tuple(zip(self.keys(), self.values()))

    # def keys(self) -> tuple[Hashable, ...]: # type: ignore
    #     """Returns 'contents' keys as a tuple.
        
    #     Returns:
    #         tuple[Hashable, ...]: a tuple equivalent to dict.keys().
            
    #     """
    #     return tuple(self.contents.keys())

    def setdefault(self, value: Any) -> None: # type: ignore
        """sets default value to return when 'get' method is used.
        
        Args:
            value (Any): default value to return when 'get' is called and the
                'default' parameter to 'get' is None.
            
        """
        self.default_factory = value 
        return
               
    def subset(
        self, 
        include: Optional[Union[Hashable, Sequence[Hashable]]] = None, 
        exclude: Optional[Union[Hashable, Sequence[Hashable]]] = None) -> (
            Dictionary):
        """Returns a new instance with a subset of 'contents'.

        This method applies 'include' before 'exclude' if both are passed. If
        'include' is None, all existing keys will be added before 'exclude' is
        applied.
        
        Args:
            include (Optional[Union[Hashable, Sequence[Hashable]]]): key(s) to 
                include in the new Dictionary instance.
            exclude (Optional[Union[Hashable, Sequence[Hashable]]]): key(s) to 
                exclude in the new Dictionary instance.                
        
        Raises:
            ValueError: if 'include' and 'exclude' are both None.
                  
        Returns:
            Dictionary: with only keys from 'include' and no keys in 'exclude'.

        """
        if include is None and exclude is None:
            raise ValueError('include or exclude must not be None')
        else:
            if include is None:
                contents = self.contents
            else:
                include = list(convert.iterify(item = include)) 
                contents = {k: self.contents[k] for k in include}
            if exclude is not None:
                exclude = list(convert.iterify(item = exclude))
                contents = {
                    k: v for k, v in contents.items() 
                    if k not in exclude}
            new_dictionary = copy.deepcopy(self)
            new_dictionary.contents = contents
        return new_dictionary
      
    # def values(self) -> tuple[Any, ...]: # type: ignore
    #     """Returns 'contents' values as a tuple.
        
    #     Returns:
    #         tuple[Any, ...]: a tuple equivalent to dict.values().
            
    #     """
    #     return tuple(self.contents.values())

    """ Dunder Methods """

    def __getitem__(self, key: Hashable) -> Any:
        """Returns value for 'key' in 'contents'.

        Args:
            key (Hashable): key in 'contents' for which a value is sought.

        Returns:
            Any: value stored in 'contents'.

        """
        return self.contents[key]

    def __setitem__(self, key: Hashable, value: Any) -> None:
        """sets 'key' in 'contents' to 'value'.

        Args:
            key (Hashable): key to set in 'contents'.
            value (Any): value to be paired with 'key' in 'contents'.

        """
        self.contents[key] = value
        return

    def __delitem__(self, key: Hashable) -> None:
        """Deletes 'key' in 'contents'.

        Args:
            key (Hashable): key in 'contents' to delete the key/value pair.

        """
        del self.contents[key]
        return


@dataclasses.dataclass  # type: ignore
class Catalog(Dictionary):
    """Wildcard and list-accepting dictionary.

    A Catalog inherits the differences between a Dictionary and an ordinary 
    python dict.

    A Catalog differs from a Dictionary in 5 significant ways:
        1) It recognizes an 'all' key which will return a list of all values
            stored in a Catalog instance.
        2) It recognizes a 'default' key which will return all values matching
            keys listed in the 'default' attribute. 'default' can also be set
            using the 'catalog['default'] = new_default' assignment. If 
            'default' is not passed when the instance is initialized, the 
            initial value of 'default' is 'all'.
        3) It recognizes a 'none' key which will return an empty list.
        4) It supports a list of keys being accessed with the matching values 
            returned. For example, 'catalog[['first_key', 'second_key']]' will 
            return the values for those keys in a list ['first_value',
            'second_value'].
        5) If a single key is sought, a Catalog can either return the stored
            value or a stored value in a list (if 'always_return_list' is
            True). The latter option is available to make iteration easier
            when the iterator assumes a single type will be returned.

    Args:
        contents (Mapping[Hashable, Any]]): stored dictionary. Defaults to an 
            empty dict.
        default_factory (Any): default value to return when the 'get' method is 
            used.
        default (Sequence[Any]]): a list of keys in 'contents' which will be 
            used to return items when 'default' is sought. If not passed, 
            'default' will be set to all keys.
        always_return_list (bool): whether to return a list even when the key 
            passed is not a list or special access key (True) or to return a 
            list only when a list or special access key is used (False). 
            Defaults to False.
                     
    """
    contents: Mapping[Hashable, Any] = dataclasses.field(
        default_factory = dict)
    default_factory: Optional[Any] = None
    default: Optional[Any] = 'all'
    always_return_list: bool = False

    """ Dunder Methods """

    def __getitem__(
        self, 
        key: Union[Hashable, Sequence[Hashable]]) -> Union[Any, Sequence[Any]]:
        """Returns value(s) for 'key' in 'contents'.

        The method searches for 'all', 'default', and 'none' matching wildcard
        options before searching for direct matches in 'contents'.

        Args:
            key (Union[Hashable, Sequence[Hashable]]): key(s) in 'contents'.

        Returns:
            Union[Any, Sequence[Any]]: value(s) stored in 'contents'.

        """
        # Returns a list of all values if the 'all' key is sought.
        if key in _ALL_KEYS:
            return list(self.contents.values())
        # Returns a list of values for keys listed in 'default' attribute.
        elif key in _DEFAULT_KEYS:
            return self[self.default]
        # Returns an empty list if a null value is sought.
        elif key in _NONE_KEYS:
            if self.default_factory is None:
                if self.always_return_list:
                    return []
                else:
                    return None
            else:
                try:
                    return self.default_factory()
                except TypeError:
                    return self.default_factory
        # Returns list of matching values if 'key' is list-like.        
        elif isinstance(key, Sequence) and not isinstance(key, str):
            return [self.contents[k] for k in key if k in self.contents]
        # Returns matching value if key is not a non-str Sequence or wildcard.
        else:
            try:
                if self.always_return_list:
                    return [self.contents[key]]
                else:
                    return self.contents[key]
            except KeyError:
                raise KeyError(f'{key} is not in {self.__class__.__name__}')

    def __setitem__(
        self, 
        key: Union[Hashable, Sequence[Hashable]], 
        value: Union[Any, Sequence[Any]]) -> None:
        """sets 'key' in 'contents' to 'value'.

        Args:
            key (Union[Hashable, Sequence[Hashable]]): key(s) to set in 
                'contents'.
            value (Union[Any, Sequence[Any]]): value(s) to be paired with 'key' 
                in 'contents'.

        """
        try:
            self.contents[key] = value
        except TypeError:
            self.contents.update(dict(zip(key, value))) # type: ignore
        return

    def __delitem__(self, key: Union[Hashable, Sequence[Hashable]]) -> None:
        """Deletes 'key' in 'contents'.

        Args:
            key (Union[Hashable, Sequence[Hashable]]): name(s) of key(s) in 
                'contents' to delete the key/value pair.

        """
        keys = list(convert.iterify(item = key))
        if all(k in self for k in keys):
            self.contents = {
                i: self.contents[i] for i in self.contents if i not in keys}
        else:
            raise KeyError(f'{key} not found in the Catalog')
        return


@dataclasses.dataclass  # type: ignore
class Library(MutableMapping):
    """Stores classes and class instances.
    
    When searching for matches, instances are prioritized over classes.
    
    Args:
        classes (Catalog): a catalog of stored classes. Defaults to any empty
            Catalog.
        instances (Catalog): a catalog of stored class instances. Defaults to an
            empty Catalog.
                 
    """
    classes: Catalog[str, Type[Any]] = dataclasses.field(
        default_factory = Catalog)
    instances: Catalog[str, object] = dataclasses.field(
        default_factory = Catalog)
        
    """ Public Methods """
    
    def deposit(
        self, 
        item: Union[Type[Any], object],
        name: Optional[Hashable] = None) -> None:
        """Adds 'item' to 'classes' and/or 'instances'.

        If 'item' is a class, it is added to 'classes.' If it is an object, it
        is added to 'instances' and its class is added to 'classes'.
        
        Args:
            item (Union[Type, object]): class or instance to add to the Library
                instance.
            name (Optional[Hashable]): key to use to store 'item'. If not
                passed, a key will be created using the 'get_name' method.
                
        """
        key = name or traits.get_name(item = item)
        if inspect.isclass(item):
            self.classes[key] = item
        elif isinstance(item, object):
            self.instances[key] = item
            self.deposit(item = item.__class__)
        else:
            raise TypeError(f'item must be a class or a class instance')
        return
    
    def remove(self, item: Hashable) -> None:
        """Removes an item from 'instances' or 'classes.'
        
        If 'item' is found in 'instances', it will not also be removed from 
        'classes'.

        Args:
            item (Hashable): key name of item to remove.
            
        Raises:
            KeyError: if 'item' is neither found in 'instances' or 'classes'.

        """
        try:
            del self.instances[item]
        except KeyError:
            try:
                del self.classes[item]
            except KeyError:
                raise KeyError(f'{item} is not found in the Library')
        return    

    def withdraw(
        self, 
        item: Union[Hashable, Sequence[Hashable]], 
        parameters: Optional[MutableMapping[Hashable, Any]] = None) -> (
            Union[Type[Any], object]):
        """Returns instance or class of first match of 'item' from catalogs.
        
        The method prioritizes the 'instances' catalog over 'classes' and any
        passed names in the order they are listed.
        
        Args:
            item (Union[Hashable, Sequence[Hashable]]): key name(s) of stored 
                item(s) sought.
            parameters (Optional[MutableMapping[Hashable, Any]]]): keyword 
                arguments to pass to a newly created instance or, if the stored 
                item is already an instance to be manually added as attributes. 
                If not passed, the found item will be returned unaltered. 
                Defaults to None.
            
        Raises:
            KeyError: if 'item' does not match a key to a stored item in either
                'instances' or 'classes'.
            
        Returns:
            Union[Type[Any], object]: returns a class or instance if 'parameters' 
                are None, depending upon with Catalog the matching item is 
                found. If 'parameters' are passed, an instance is always returned.
            
        """
        items = list(convert.iterify(item))
        item = None
        for key in items:
            for catalog in ['instances', 'classes']:
                try:
                    item = getattr(self, catalog)[key]
                    break
                except KeyError:
                    pass
            if item is not None:
                break
        if item is None:
            raise KeyError(f'No matching item for {item} was found')
        if parameters is not None:
            if ('name' in item.__annotations__.keys() 
                    and 'name' not in parameters):
                parameters['name'] = items[0]
            if inspect.isclass(item):
                return item(**parameters)
            else:
                instance = copy.deepcopy(item)
                for key, value in parameters.items():
                    setattr(instance, key, value)
                return instance
        return item # type: ignore
    
    """ Dunder Methods """

    def __getitem__(self, key: Hashable) -> Any:
        """Returns value for 'key' in 'contents'.

        Args:
            key (Hashable): key in 'contents' for which a value is sought.

        Returns:
            Any: value stored in 'contents'.

        """
        return self.withdraw(item = key)

    def __setitem__(self, key: Hashable, value: Any) -> None:
        """sets 'key' in 'contents' to 'value'.

        Args:
            key (Hashable): key to set in 'contents'.
            value (Any): value to be paired with 'key' in 'contents'.

        """
        self.deposit(item = value, name = key)
        return

    def __delitem__(self, key: Hashable) -> None:
        """Deletes 'key' in 'contents'.

        Args:
            key (Hashable): key in 'contents' to delete the key/value pair.

        """
        self.remove(item = key)
        return

    def __iter__(self) -> Iterator[Any]:
        """Returns iterable of 'contents'.

        Returns:
            Iterator: of 'contents'.

        """
        combined = copy.deepcopy(self.instances)
        return iter(combined.update(self.classes))

    def __len__(self) -> int:
        """Returns length of 'contents'.

        Returns:
            int: length of 'contents'.

        """
        combined = copy.deepcopy(self.instances)
        return len(combined.update(self.classes))
    


"""
composites: base classes for extensible, flexible, lightweight data structures
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2021, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:

          
To Do:
    Integrate ashford Kinds system when it is finished.
    Add in 'beautify' str representations from amos once those are finished.
    Add universal 'merge' method in Composite to replace the ad hoc methods of
        subclasses. This requires completion of "To Do:" tasks in the 'convert'
        module as well.
    
"""
from __future__ import annotations
import abc
from collections.abc import Collection, Hashable, MutableMapping, Sequence
import dataclasses
import inspect
from typing import Any, ClassVar, Optional, Type, TYPE_CHECKING, Union

from ..base import bunches
from ..construct import factories
from ..observe import traits
from . import check


if TYPE_CHECKING:
    from . import hybrid

   
@dataclasses.dataclass
class Node(object):
    """Vertex wrapper to provide hashability to any object.
    
    Node acts a basic wrapper for any item stored in a composite structure.
    
    Args:
        contents (Optional[Any]): any stored item(s). Defaults to None.
        name (Optional[str]): designates the name of a class instance that is 
            used for internal and external referencing in a composite object.
            Defaults to None.
            
    """
    contents: Optional[Any] = None
    name: Optional[str] = None

    """ Initialization Methods """
    
    def __init_subclass__(cls, *args: Any, **kwargs: Any):
        """Forces subclasses to use the same hash methods as Node.
        
        This is necessary because dataclasses, by design, do not automatically 
        inherit the hash and equivalance dunder methods from their super 
        classes.
        
        """
        # Calls other '__init_subclass__' methods for parent and mixin classes.
        try:
            super().__init_subclass__(*args, **kwargs) # type: ignore
        except AttributeError:
            pass
        # Copies hashing related methods to a subclass.
        cls.__hash__ = Node.__hash__ # type: ignore
        cls.__eq__ = Node.__eq__ # type: ignore
        cls.__ne__ = Node.__ne__ # type: ignore

    def __post_init__(self) -> None:
        """Initializes class instance attributes."""
        # Sets 'name' attribute if 'name' is None.
        self.name = self.name or traits.get_name(item = self)
                
    """ Dunder Methods """
    
    @classmethod
    def __subclasshook__(cls, subclass: Type[Any]) -> bool:
        """Returns whether 'subclass' is a virtual or real subclass.

        Args:
            subclass (Type[Any]): item to test as a subclass.

        Returns:
            bool: whether 'subclass' is a real or virtual subclass.
            
        """
        return issubclass(subclass, Hashable) and hasattr(subclass, 'name')
               
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        """Returns whether 'instances' is an instance of this class.

        Args:
            instance (object): item to test as an instance.

        Returns:
            bool: whether 'instance' is an instance of this class.
            
        """
        return check.is_node(item = instance)
    
    def __hash__(self) -> int:
        """Makes Node hashable so that it can be used as a key in a dict.

        Rather than using the object ID, this method prevents two Nodes with
        the same name from being used in a composite object that uses a dict as
        its base storage type.
        
        Returns:
            int: hashable of 'name'.
            
        """
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        """Makes Node hashable so that it can be used as a key in a dict.

        Args:
            other (object): other object to test for equivalance.
            
        Returns:
            bool: whether 'name' is the same as 'other.name'.
            
        """
        try:
            return str(self.name) == str(other.name) # type: ignore
        except AttributeError:
            return str(self.name) == other

    def __ne__(self, other: object) -> bool:
        """Completes equality test dunder methods.

        Args:
            other (object): other object to test for equivalance.
           
        Returns:
            bool: whether 'name' is not the same as 'other.name'.
            
        """
        return not(self == other)

    def __contains__(self, item: Any) -> bool:
        """Returns whether 'item' is in or equal to 'contents'.

        Args:
            item (Any): item to check versus 'contents'
            
        Returns:
            bool: if 'item' is in or equal to 'contents' (True). Otherwise, it
                returns False.

        """
        try:
            return item in self.contents
        except TypeError:
            try:
                return item is self.contents
            except TypeError:
                return item == self.contents
                
    def __getattr__(self, attribute: str) -> Any:
        """Looks for 'attribute' in 'contents'.

        Args:
            attribute (str): name of attribute to return.

        Raises:
            AttributeError: if 'attribute' is not found in 'contents'.

        Returns:
            Any: matching attribute.

        """
        try:
            return object.__getattribute__(
                object.__getattribute__(self, 'contents'), attribute)
        except AttributeError:
            raise AttributeError(
                f'{attribute} was not found in '
                f'{object.__getattribute__(self, "__name__")}') 

    def __setattr__(self, attribute: str, value: Any) -> None:
        """Sets 'attribute' to 'value'.
        
        If 'attribute' exists in this class instance, its new value is set to
        'value.' Otherwise, 'attribute' and 'value' are set in what is stored
        in 'contents'

        Args:
            attribute (str): name of attribute to set.
            value (Any): value to store in 'attribute'.

        """
        if hasattr(self, attribute) or self.contents is None:
            object.__setattr__(self, attribute, value)
        else:
            object.__setattr__(self.contents, attribute, value)
            
    def __delattr__(self, attribute: str) -> None:
        """Deletes 'attribute'.
        
        If 'attribute' exists in this class instance, it is deleted. Otherwise, 
        this method attempts to delete 'attribute' from what is stored in 
        'contents'

        Args:
            attribute (str): name of attribute to set.

        Raises:
            AttributeError: if 'attribute' is neither found in a class instance
                nor in 'contents'.
            
        """
        try:
            object.__delattr__(self, attribute)
        except AttributeError:
            try:
                object.__delattr__(self.contents, attribute)
            except AttributeError:
                raise AttributeError(
                    f'{attribute} was not found in '
                    f'{object.__getattribute__(self, "__name__")}') 

 
@dataclasses.dataclass
class Nodes(bunches.Bunch, abc.ABC):
    """Collection of nodes.
    
    Nodes are not guaranteed to be in order. 

    Args:
        contents (Optional[Any]): any stored item(s). Defaults to None.
            
    """
    contents: Optional[Collection[Node]] = None
    
    """ Dunder Methods """ 
    
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        return check.is_nodes(item = instance)
    

def nodify(item: Union[Type[Any], object]) -> Union[Type[Node], Node]:
    """Converts a class or object into a node for a composite data structure.
    
    Currently, the method supports any object but only python dataclass types 
    for classes. And those dataclasses should not have a '__post_init__' 
    method - it will be overwritten by 'nodify'.

    Args:
        item (Union[Type[Any], object]): class or instance to transform into a  
            node.

    Returns:
        Union[Type[Node], Node]: a node class or instance.
        
    """
    item.__hash__ = Node.__hash__ # type: ignore
    item.__eq__ = Node.__eq__ # type: ignore
    item.__ne__ = Node.__ne__ # type: ignore
    if inspect.isclass(item):
        item.__post_init__ = Node.__post_init__ # type: ignore
    else:
        if not hasattr(item, 'name') or not item.name:
            item.name = traits.get_name(item = item)
    return item

           
@dataclasses.dataclass
class Composite(factories.RegistrarFactory, abc.ABC):
    """Base class for composite data structures.
    
    Args:
        contents (Collection[Any]): stored collection of nodes and/or edges.
        registry (ClassVar[MutableMapping[str, Type[Any]]]): key names are str
            names of a subclass (snake_case by default) and values are the 
            subclasses. Defaults to an empty dict.  
                                     
    """  
    contents: Collection[Any]
    registry: ClassVar[MutableMapping[str, Type[Any]]] = {}
    
    """ Required Subclass Properties """
        
    @abc.abstractproperty
    def endpoint(self) -> Optional[Union[Node, Nodes]]:
        """Returns the endpoint(s) of the stored composite object."""
        pass
 
    @abc.abstractproperty
    def root(self) -> Optional[Union[Node, Nodes]]:
        """Returns the root(s) of the stored composite object."""
        pass

    # @abc.abstractproperty
    # def adjacency(self) -> graph.Adjacency:
    #     """Returns the stored composite object as an graph.Adjacency."""
    #     pass

    # @abc.abstractproperty
    # def edges(self) -> graph.Edges:
    #     """Returns the stored composite object as an graph.Edges."""
    #     pass
       
    # @abc.abstractproperty
    # def matrix(self) -> graph.Matrix:
    #     """Returns the stored composite object as a graph.Matrix."""
    #     pass
       
    # @abc.abstractproperty
    # def nodes(self) -> Nodes:
    #     """Returns the stored composite object as a Nodes."""
    #     pass
        
    # @abc.abstractproperty
    # def pipeline(self) -> hybrid.Pipeline:
    #     """Returns the stored composite object as a hybrid.Pipeline."""
    #     pass
        
    # @abc.abstractproperty
    # def pipelines(self) -> hybrid.Pipelines:
    #     """Returns the stored composite object as a hybrid.Pipelines."""
    #     pass
            
    # @abc.abstractproperty
    # def tree(self) -> tree.Tree:
    #     """Returns the stored composite object as a tree.Tree."""
    #     pass
                 
    # """ Required Subclass Class Methods """
    
    # @abc.abstractclassmethod
    # def from_adjacency(cls, item: graph.Adjacency) -> Composite:
    #     """Creates a Composite instance from an graph.Adjacency."""
    #     pass
    
    # @abc.abstractclassmethod
    # def from_edges(cls, item: graph.Edges) -> Composite:
    #     """Creates a Composite instance from an graph.Edges."""
    #     pass
    
    # @abc.abstractclassmethod
    # def from_matrix(cls, item: graph.Matrix) -> Composite:
    #     """Creates a Composite instance from a graph.Matrix."""
    #     pass
    
    # @abc.abstractclassmethod
    # def from_pipeline(cls, item: hybrid.Pipeline) -> Composite:
    #     """Creates a Composite instance from a hybrid.Pipeline."""
    #     pass
    
    # @abc.abstractclassmethod
    # def from_pipelines(cls, item: hybrid.Pipelines) -> Composite:
    #     """Creates a Composite instance from a hybrid.Pipelines."""
    #     pass

    # @abc.abstractclassmethod
    # def from_tree(cls, item: tree.Tree) -> Composite:
    #     """Creates a Composite instance from a tree.Tree."""
    #     pass
                 
    """ Required Subclass Methods """
    
    @abc.abstractmethod
    def add(self, item: Node, *args: Any, **kwargs: Any) -> None:
        """Adds 'node' to the stored composite object.
        
        Args:
            node (Node): a node to add to the stored composite object.
                
        """
        pass
    
    @abc.abstractmethod
    def append(
        self, 
        item: Union[Node, Nodes, Composite], 
        *args: Any, 
        **kwargs: Any) -> None:
        """Appends 'item' to the endpoint(s) of the stored composite object.

        Args:
            item (Union[Node, Composite]): a single Node or other Composite
                object to add to the stored composite object.
                
        """
        pass
        
    @abc.abstractmethod
    def delete(self, item: Any, *args: Any, **kwargs: Any) -> None:
        """Deletes node from the stored composite object.
        
        Args:
            item (Any): node or key to the a node to delete.
        
        Raises:
            KeyError: if 'item' is not in 'contents'.
            
        """
        pass
  
    @abc.abstractmethod
    def merge(self, item: Composite, *args: Any, **kwargs: Any) -> None:
        """Combines 'item' with the stored composite object.

        Args:
            item (Composite): another Composite object to add to the stored 
                composite object.
                
        """
        pass
    
    @abc.abstractmethod
    def prepend(
        self, 
        item: Union[Node, Nodes, Composite], 
        *args: Any, 
        **kwargs: Any) -> None:
        """Prepends 'item' to the root(s) of the stored composite object.

        Args:
            item (Union[Node, Composite]): a single Node or other Composite
                object to add to the stored composite object.
                
        """
        pass
    
    @abc.abstractmethod
    def subset(
        self, 
        include: Union[Any, Sequence[Any]] = None,
        exclude: Union[Any, Sequence[Any]] = None, 
        *args: Any, 
        **kwargs: Any) -> Composite:
        """Returns a new Composite with a subset of 'contents'.
        
        Args:
            include (Union[Any, Sequence[Any]]): nodes which should be included
                in the new Composite.
            exclude (Union[Any, Sequence[Any]]): nodes which should not be 
                in the new Composite.

        Returns:
           Composite: with only nodes indicated by 'include' and 'exclude'.
           
        """
        pass
    
    @abc.abstractmethod
    def walk(
        self, 
        start: Optional[Node] = None,
        stop: Optional[Node] = None, 
        path: Optional[hybrid.Pipeline] = None,
        return_pipelines: bool = True, 
        *args: Any, 
        **kwargs: Any) -> Union[hybrid.Pipeline, hybrid.Pipelines]:
        """Returns path in the stored composite object from 'start' to 'stop'.
        
        Args:
            start (Optional[Node]): node to start paths from. Defaults to None.
                If it is None, 'start' should be assigned to one of the roots
                of the Composite.
            stop (Optional[Node]): node to stop paths. Defaults to None. If it 
                is None, 'start' should be assigned to one of the roots of the 
                Composite.
            path (Optional[hybrid.Pipeline]): a path from 'start' to 'stop'. 
                Defaults to None. This parameter is used by recursive methods 
                for determining a path.
            return_pipelines (bool): whether to return a Pipelines instance 
                (True) or a hybrid.Pipeline instance (False). Defaults to True.

        Returns:
            Union[hybrid.Pipeline, hybrid.Pipelines]: path(s) through the 
                Composite object. If multiple paths are possible and 
                'return_pipelines' is False, this method should return a 
                Pipeline that includes all such paths appended to each other. If 
                multiple paths are possible and 'return_pipelines' is True, a 
                Pipelines instance with all of the paths should be returned. 
                Defaults to True.
                            
        """
        pass
    
    """ Dunder Methods """

    def __add__(self, other: Composite) -> None:
        """Adds 'other' to the stored composite object using 'append'.

        Args:
            other (Union[base.Composite]): another Composite subclass instance 
                to add to the current one.
            
        """
        self.append(item = other)     
        return 

    def __radd__(self, other: Composite) -> None:
        """Adds 'other' to the stored composite object using 'prepend'.

        Args:
            other (Union[base.Composite]): another Composite subclass instance 
                to add to the current one.
            
        """
        self.prepend(item = other)     
        return 

    # def __str__(self) -> str:
    #     """Returns prettier str representation of the stored graph.

    #     Returns:
    #         str: a formatted str of class information and the contained graph.
            
    #     """
    #     return amos.recap.beautify(item = self, package = 'piles')  




"""
factories: factory creation tools
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2021, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:   

        
ToDo:


"""
from __future__ import annotations
import abc
from collections.abc import Mapping, MutableMapping, Sequence
import dataclasses
from typing import Any, ClassVar, Optional, Type, Union

from ..base import mappings
from ..observe import registries
from ..observe import traits
from ..repair import modify

 
@dataclasses.dataclass
class BaseFactory(abc.ABC):
    """Base class for instance factory mixins."""
    
    """ Required Subclass Methods """

    @abc.abstractclassmethod
    def create(cls, item: Any, **kwargs: Any) -> BaseFactory:
        """Implements technique for creating a (sub)class instance.

        Args:
            item (Any): argument indicating creation method to use.

        Returns:
            BaseFactory: instance of a SourcesFactory.
            
        """
        pass

   
@dataclasses.dataclass
class LibraryFactory(BaseFactory):
    """Mixin which automatically registers subclasses and instances.
    
    Args:
        library (ClassVar[mappings.Library]): project library of classes, 
            instances, and base classes. 
            
    """
    library: ClassVar[mappings.Library] = mappings.Library()
    
    """ Initialization Methods """
    
    @classmethod
    def __init_subclass__(cls, *args: Any, **kwargs: Any):
        """Automatically registers subclass."""
        # Because LibraryFactory is used as a mixin, it is important to
        # call other base class '__init_subclass__' methods, if they exist.
        try:
            super().__init_subclass__(*args, **kwargs) # type: ignore
        except AttributeError:
            pass
        name = traits.get_name(item = cls)
        cls.library.deposit(item = cls, name = name)
            
    def __post_init__(self) -> None:
        try:
            super().__post_init__(*args, **kwargs) # type: ignore
        except AttributeError:
            pass
        key = traits.get_name(item = self)
        self.__class__.library.deposit(item = self, name = key)
    
    """ Public Methods """

    @classmethod
    def create(
        cls, 
        item: Union[str, Sequence[str]], 
        **kwargs: Any) -> LibraryFactory:
        """Creates an instance of a LibraryFactory subclass from 'item'.
        
        Args:
            item (Any): any supported data structure which acts as a source for
                creating a LibraryFactory or a str which matches a key in 
                'library'.
                                
        Returns:
            LibraryFactory: a LibraryFactory subclass instance created based 
                on 'item' and any passed arguments.
                
        """
        return cls.library.withdraw(item, **kwargs)

             
@dataclasses.dataclass
class RegistrarFactory(registries.Registrar, abc.ABC):
    """Mixin which automatically registers subclasses for use by a factory.
    
    Args:
        registry (ClassVar[MutableMapping[str, Type[Any]]]): key names are str
            names of a subclass (snake_case by default) and values are the 
            subclasses. Defaults to an empty dict.  
            
    """
    registry: ClassVar[MutableMapping[str, Type[Any]]] = {}
    
    """ Public Methods """

    @classmethod
    def create(cls, item: Any, **kwargs: Any) -> RegistrarFactory:
        """Creates an instance of a RegistrarFactory subclass from 'item'.
        
        Args:
            item (Any): any supported data structure which acts as a source for
                creating a RegistrarFactory or a str which matches a key in 
                'registry'.
                                
        Returns:
            RegistrarFactory: a RegistrarFactory subclass instance created based 
                on 'item' and any passed arguments.
                
        """
        if isinstance(item, str):
            try:
                return cls.registry[item](**kwargs)
            except KeyError:
                pass
        try:
            name = traits.get_name(item = item)
            return cls.registry[name](item, **kwargs)
        except KeyError:
            for name, kind in cls.registry.items():
                if (
                    abc.ABC not in kind.__bases__ 
                    and kind.__instancecheck__(instance = item)):
                    method = getattr(cls, f'from_{name}')
                    return method(item, **kwargs)       
            raise ValueError(
                f'Could not create {cls.__name__} from item because it '
                f'is not one of these supported types: '
                f'{str(list(cls.registry.keys()))}')
            
                      
@dataclasses.dataclass
class SourcesFactory(BaseFactory, abc.ABC):
    """Supports subclass creation using 'sources' class attribute.

    Args:
        sources (str, str]]): keys are str names of the types of the data 
            sources for object creation. For the appropriate creation 
            classmethod to be called, the types need to match the type of the
            first argument passed.  
    
    """
    sources: ClassVar[Mapping[str, str]] = {}
    
    """ Public Methods """

    @classmethod
    def create(cls, item: Any, **kwargs: Any) -> SourcesFactory:
        """Calls corresponding creation class method to instance a class.
        
        For create to work properly, there should be a corresponding classmethod
        named f'from_{value in sources}'. If you would prefer a different naming
        format, you can subclass SourcesFactory and override the 
        '_get_create_method_name' classmethod.

        Raises:
            AttributeError: If an appropriate method does not exist for the
                data type of 'item.'
            KeyError: If the type of 'item' does not match a key in 
                'sources'.

        Returns:
            TypeFactory: instance of a SourcesFactory.
            
        """
        for kind, suffix in cls.sources.items():
            if isinstance(item, kind):
                method_name = cls._get_create_method_name(item = suffix)
                try:
                    method = getattr(cls, method_name)
                except AttributeError:
                    raise AttributeError(f'{method_name} does not exist')
                return method(item, **kwargs)
        raise KeyError(
            f'item does not match any recognized types in sources attribute')

    """ Private Methods """
    
    @classmethod
    def _get_create_method_name(cls, item: str) -> str:
        """Returns classmethod name for creating an instance.
        
        Args:
            item (str): name corresponding to part of the str of the method
                name used for instancing.
                
        """
        return f'from_{item}'
      

@dataclasses.dataclass
class SubclassFactory(BaseFactory, abc.ABC):
    """Returns a subclass based on arguments passed to the 'create' method."""
        
    """ Public Methods """

    @classmethod
    def create(cls, item: str, **kwargs: Any) -> SubclassFactory:
        """Returns subclass based on 'item'
        
        A subclass in the '__subclasses__' attribute is selected based on the
        snake-case name of the subclass.
        
        Raises:
            KeyError: If a corresponding subclass does not exist for 'item.'

        Returns:
            SubclassFactory: instance of a SubclassFactory subclass.
            
        """
        options = {
            modify.snakify(item = s.__name__): s for s in cls.__subclasses__}
        try:
            return options[item](**kwargs)
        except KeyError:
            raise KeyError(f'No subclass {item} was found')
                 
          
@dataclasses.dataclass
class TypeFactory(BaseFactory, abc.ABC):
    """Supports subclass creation using str name of item type passed."""
    
    """ Public Methods """

    @classmethod
    def create(cls, item: Any, **kwargs: Any) -> TypeFactory:
        """Calls construction method based on type of 'item'.
        
        For create to work properly, there should be a corresponding classmethod
        named f'from_{snake-case str name of type}'. If you would prefer a 
        different naming format, you can subclass TypeFactory and override the 
        '_get_create_method_name' classmethod.

        Raises:
            AttributeError: If an appropriate method does not exist for the
                data type of 'item.'

        Returns:
            TypeFactory: instance of a TypeFactory.
            
        """
        suffix = modify.snakify(item = str(type(item)))
        method_name = cls._get_create_method_name(item = suffix)
        try:
            method = getattr(cls, method_name)
        except AttributeError:
            raise AttributeError(f'{method_name} does not exist')
        return method(item, **kwargs)

    """ Private Methods """
    
    @classmethod
    def _get_create_method_name(cls, item: str) -> str:
        """Returns classmethod name for creating an instance.
        
        Args:
            item (str): name corresponding to part of the str of the method
                name used for instancing.
                
        """
        return f'from_{item}'
